# peppy Copyright (c) 2006-2010 Rob McMullen
# Licenced under the GPLv2; see http://peppy.flipturn.org for more info
import os, re, time, codecs

import wx
import wx.stc

from cStringIO import StringIO

from stcinterface import *
from omnivore.utils.textutil import *
from omnivore.utils.clipboard import *

import logging
log = logging.getLogger(__name__)


class FoldNode:
    def __init__(self,level,start,end,text,parent=None,styles=[]):
        """Folding node as data for tree item."""
        self.parent     = parent
        self.level      = level
        self.start      = start
        self.end        = end
        self.text       = text
        self.styles     = styles #can be useful for icon detection
        self.children   = []

    def __str__(self):
        return "L%d s%d e%d %s" % (self.level, self.start, self.end, self.text.rstrip())


class PeppyBaseSTC(wx.stc.StyledTextCtrl, STCInterface):
    """The base class used as backend storage for files that fit in memory.
    
    Peppy uses the wx.stc.StyledTextCtrl (abbreviated STC in the peppy
    documentation) to store its files in memory.  Because the STC allows
    multiple views of the same file, we take advantage of this to prevent
    keeping multiple copies of the file for different views.  As described
    in L{MajorMode}, multiple STCs will keep themselves updated when the user
    makes change in another STC.
    
    Note that there is a difference in the STC stored in the L{Buffer} and any
    STCs that are used by the L{FundamentalMode} class.  There will always be
    at least two instances of STCs: one in the buffer that is the main backend
    storage, and the other in the major mode that provides the view.  They
    both point to the same data, but they are kept separate in order that
    major modes (i.e.  the views of the data) can change at will.
    
    This class performs the bookkeeping to keep the STC document pointers up to
    date when a new view is added.
    """
    eol2int = {'\r': wx.stc.STC_EOL_CR,
               '\r\n': wx.stc.STC_EOL_CRLF,
               '\n': wx.stc.STC_EOL_LF,
               }
    int2eol = {wx.stc.STC_EOL_CR: '\r',
               wx.stc.STC_EOL_CRLF: '\r\n',
               wx.stc.STC_EOL_LF: '\n',
               }

    def __init__(self, parent, refstc=None, copy=None, **kwargs):
        """Initialize the styled text control with peppy extensions.
        
        @param parent: wx.Control used as the parent control
        
        @keyword refstc: (default None) the parent STC, if any, to which
        this document object is linked.  Edits in one document are reflected
        transparently in all the linked documents.
        
        @keyword copy: (default None) STC from which to copy the text into the
        body of this STC
        """
        wx.stc.StyledTextCtrl.__init__(self, parent, -1, pos=(9000,9000), **kwargs)
        self.ClearAll()

        if refstc is not None:
            self.refstc = refstc
            self.docptr = self.refstc.docptr
            self.AddRefDocument(self.docptr)
            self.SetDocPointer(self.docptr)
            self.SetCodePage(65001) # set for unicode character display
            self.refstc.addSubordinate(self)
            log.debug("referencing document %s" % self.docptr)
        else:
            self.refstc = self
            self.docptr = self.CreateDocument()
            self.SetDocPointer(self.docptr)
            self.SetCodePage(65001) # set for unicode character display
            self.encoding = None # we don't know the encoding yet
            self.bom = None # we don't know if there is a Byte Order Mark
            log.debug("creating new document %s" % self.docptr)
            self.subordinates = []

            # If views can share info among similar classes, that info can be
            # stored here.
            self.stc_classes = []
            self.stc_class_info = {}

            if copy is not None:
                txt = copy.GetStyledText(0,copy.GetTextLength())
                dprint("copying %s from old stc." % repr(txt))
                self.AddStyledText(txt)
        self.maybe_undo_eolmode = None

    def updateSubordinateClasses(self):
        """Update the list of classes viewing this buffer."""
        classes = {}
        for view in self.subordinates:
            classes[view.__class__] = True
        self.stc_classses = classes.keys()

    def getSharedClassInfo(self, cls):
        """Get the dict that can be used to store data common to viewers of
        the specified class.
        """
        if cls not in self.refstc.stc_class_info:
            self.refstc.stc_class_info[cls] = {}
        return self.refstc.stc_class_info[cls]

    def addSubordinate(self,otherstc):
        self.subordinates.append(otherstc)
        self.updateSubordinateClasses()

    def removeDocumentView(self):
        """Remove the reference of this view from its parent document"""
        if self.refstc != self:
            self.refstc.subordinates.remove(self)
            self.refstc.updateSubordinateClasses()

    def open(self, buffer, message=None):
        """Open the file from the buffer and read in its contents.
        
        The open method is provided here in the STC and not in the buffer so
        that specialized STCs can provide their own load methods.  So far,
        this is only used in the L{NonResidentSTC} for files that are too
        large to fit in memory.
        """
        fh = buffer.getBufferedReader()
        self.readThreaded(fh, buffer, message)

    def revertEncoding(self, buffer, url=None, message=None, encoding=None, allow_undo=False):
        if url is None:
            url = buffer.url
        fh = vfs.open(url)
        if allow_undo:
            self.BeginUndoAction()
        self.ClearAll()
        self.readThreaded(fh, buffer, message)
        self.openSuccess(buffer, encoding=encoding)
        if allow_undo:
            self.EndUndoAction()
        else:
            self.EmptyUndoBuffer()

    def readThreaded(self, fh, buffer, message=None):
        self.refstc.tempstore = StringIO()
        if fh:
            # if the file exists, read the contents.
            length = vfs.get_size(buffer.url)
            log.debug("Loading %d bytes" % length)
            chunk = 65536
            if length/chunk > 100:
                chunk *= 4
            self.readFrom(fh, chunk=chunk, length=length, message=message)
            buffer.setInitialStateIsUnmodified()
        else:
            # if the file doesn't exist, the user has asked to create a new
            # file.  Peppy needs to reflect that it hasn't been saved by
            # setting its initial state to be 'modified'
            buffer.setInitialStateIsModified()

    def openSuccess(self, buffer, headersize=1024, encoding=None):
        bytes = self.tempstore.getvalue()

        self.resetText(bytes, headersize, encoding)

        del self.tempstore

    def resetText(self, bytes, headersize=1024, encoding=None):
        numbytes = len(bytes)
        if headersize > numbytes:
            headersize = numbytes
        header = bytes[0:headersize]

        if encoding:
            # Normalize the encoding name by running it through the codecs list
            self.refstc.encoding = codecs.lookup(encoding).name
        if not self.refstc.encoding:
            self.refstc.encoding, self.refstc.bom = detectEncoding(header)
        self.decodeText(bytes)
        log.debug("found encoding = %s" % self.refstc.encoding)
        self.detectLineEndings()

    def readFrom(self, fh, amount=None, chunk=65536, length=0, message=None):
        """Read a chunk of the file from the file-like object.
        
        Rather than reading the file in with a single call to fh.read(), it is
        broken up into segments.  It may take a significant amount of time to
        read a file, either if the file is really big or the file is loaded
        over a slow URI scheme.  The threaded load capability of peppy is used
        to display a progress bar that is updated after each segment is loaded,
        and also keeps the user interface responsive during a file load.
        """
        total = 0
        while amount is None or total<amount:
            txt = fh.read(chunk)
            log.debug("reading %d bytes from %s" % (len(txt), fh))

            if len(txt) > 0:
                total += len(txt)
                if message:
                    # Negative value will switch the progress bar to
                    # pulse mode
                    #Publisher().sendMessage(message, (total*100)/length)
                    pass

                if isinstance(txt, unicode):
                    # This only seems to happen for unicode files written
                    # to the mem: filesystem, but if it does happen to be
                    # unicode, there's no need to convert the data
                    self.refstc.encoding = "utf-8"
                    self.tempstore.write(unicode.encode('utf-8'))
                else:
                    self.tempstore.write(txt)
            else:
                # stop when we reach the end.  An exception will be
                # handled outside this class
                break

    def decodeText(self, bytes):
        """Check for the file encoding and convert in place.
        
        If the encoding is embedded in the file (through emacs "magic
        comments"), change the text from the binary representation into the
        specified encoding.
        """
        if self.refstc.encoding:
            try:
                unicodestring = encodedBytesToUnicode(bytes, self.refstc.encoding, self.refstc.bom)
                log.debug("unicodestring(%s) = %s bytes" % (type(unicodestring), len(unicodestring)))
                self.SetText(unicodestring)
                return
            except UnicodeDecodeError, e:
                log.debug("bad encoding %s:" % self.refstc.encoding)
                self.refstc.badencoding = self.refstc.encoding
                self.refstc.encoding = None
                self.refstc.bom = None

        # If there's no encoding or an error in the decoding, stuff the binary
        # bytes in the stc.  The only way to load binary data into scintilla
        # is to convert it to two bytes per character: first byte is the
        # content, 2nd byte is styling (which we set to zero)
        self.SetText('')
        styledtxt = '\0'.join(bytes)+'\0'
        self.AddStyledText(styledtxt)

    def prepareEncoding(self):
        """Prepare the file for encoding.
        
        This method provides a short-circuit of the writing process in case the
        encoding is bad.  This is a poor man's way of preventing a zero-length
        file due to a bad encoding, because should this method generate an
        exception, the file will never be opened for writing and therefore
        won't be truncated.
        """
        bytes = ""
        try:
            txt = self.GetText()
            encoding, bom = detectEncoding(txt)
            if encoding:
                log.debug("found encoding %s" % encoding)
                bytes = txt.encode(encoding)
                if encoding != self.refstc.encoding:
                    # If the encoding has changed, update it here
                    self.refstc.encoding = encoding
                    self.refstc.bom = bom
                    self.decodeText(bytes)
            elif self.refstc.encoding:
                if self.refstc.bom:
                    bytes = self.refstc.bom
                bytes += txt.encode(self.refstc.encoding)
            else:
                # Have to use GetStyledText because GetText will truncate the
                # string at the first zero character.
                numchars = self.GetTextLength()
                bytes = self.GetStyledText(0, numchars)[0:numchars*2:2]

            self.refstc.encoded = bytes
        except:
            self.refstc.encoded = None
            raise

    def openFileForWriting(self, url):
        return vfs.open_write(url)

    def writeTo(self, fh, url):
        """Writes a copy of the document to the provided file-like object.
        
        Note that peppy is not currently thread-enabled during file writing.
        """
        txt = self.refstc.encoded
        if txt is None:
            raise IOError("Invalid encoded string -- this should never happen")

        #dprint("writing %d bytes to %s" % (len(txt), fh))

        try:
            fh.write(txt)
        finally:
            # clean up temporary encoded version of the text
            self.refstc.encoded = None

    def closeFileAfterWriting(self, fh):
        fh.close()

    def getAutosaveTemporaryFilename(self, buffer):
        """Hook to allow STC to override autosave filename"""
        return wx.GetApp().autosave.getFilename(buffer.url)

    def getBackupTemporaryFilename(self, buffer):
        """Hook to allow STC to override backup filename"""
        return wx.GetApp().backup.getFilename(buffer.url)

    ## Additional functionality
    def checkUndoEOL(self):
        """Check to see if the last change was converting all EOL characters.
        
        The wx.stc.StyledTextCtrl doesn't directly store if the last undo/redo
        was the change in all of the end of line characters, so we have to
        check ourselves.
        """
        # Check to see if the eol mode has changed.
        if self.maybe_undo_eolmode is not None:
            if self.maybe_undo_eolmode['likely']:
                self.detectLineEndings(self.GetText())
                #Publisher().sendMessage('resetStatusBar')
            self.maybe_undo_eolmode = None

    def Undo(self):
        """Override of base Undo command to add our additional checks."""
        wx.stc.StyledTextCtrl.Undo(self)
        self.checkUndoEOL()

    def Redo(self):
        """Override of base Redo command to add our additional checks."""
        wx.stc.StyledTextCtrl.Redo(self)
        self.checkUndoEOL()

    ## STCInterface additions
    def CanCopy(self):
        return True

    def CanCut(self):
        return True

    def SelectAll(self):
        self.SetSelectionStart(0)
        self.SetSelectionEnd(self.GetLength())

    def GetBinaryData(self, start=0, end=-1):
        """Convenience function to get binary data out of the STC.
        
        The only way to get binary data out of the STC is to use the
        GetStyledText method and chop out every other byte.  Using the regular
        GetText method will stop at the first nul character.

        @param start: first text position
        @param end: last text position
        
        @returns: binary data between start and end-1, inclusive (just
        like standard python array slicing)
        """
        if end == -1:
            end = self.GetTextLength()
        return self.GetStyledText(start,end)[::2]

    def SetBinaryData(self, loc, locend, bytes):
        """Replace the binary data in the specified range.
        
        Given a start and end position, this method replaces the binary bytes
        with the given string of bytes.  Note that bytes can be inserted or
        deleted if the length of the new byte string is different than the
        specified range.
        """
        # FIXME: the set/replace selection can fail if we start or end in the
        # middle of a multi-byte sequence.  To properly handle this, we'd
        # have to search backwards and forwards to make sure that we aren't
        # splitting a UTF-8 sequence
        start = loc
        valid = False
        while not valid:
            try:
                self.GotoPos(start)
                valid = True
            except wx._core.PyAssertionError:
                log.debug("Trying back one... start=%d" % start)
                if start > 0:
                    start -= 1
        if start > 0:
            self.CmdKeyExecute(wx.stc.STC_CMD_CHARLEFTEXTEND)
        start = self.GetSelectionStart()
        end = locend
        valid = False
        while not valid:
            try:
                self.GotoPos(end)
                valid = True
            except wx._core.PyAssertionError:
                log.debug("Trying ahead one... end=%d" % end)
                if end < self.GetLength():
                    end += 1
        self.CmdKeyExecute(wx.stc.STC_CMD_CHARRIGHTEXTEND)
        end = self.GetSelectionEnd()
        data = self.GetStyledText(start, end)
        self.SetSelection(start, end)
        self.ReplaceSelection('')

        styled = '\0'.join(bytes) + '\0'
        gap1 = loc - start
        gap2 = gap1 + locend - loc
        replacement = data[:gap1 * 2] + styled + data[gap2 * 2:]
        log.debug("start=%d loc=%d locend=%d end=%d  data=%s styled=%s replace=%s" % (start, loc, locend, end, repr(data), repr(styled), repr(replacement)))
        self.AddStyledText(replacement)

    def GuessBinary(self,amount,percentage):
        """
        Guess if the text in this file is binary or text by scanning
        through the first C{amount} characters in the file and
        checking if some C{percentage} is out of the printable ascii
        range.

        Obviously this is a poor check for unicode files, so this is
        just a bit of a hack.

        @param amount: number of characters to check at the beginning
        of the file

        @type amount: int
        
        @param percentage: percentage of characters that must be in
        the printable ASCII range

        @type percentage: number

        @rtype: boolean
        """
        endpos=self.GetLength()
        if endpos>amount: endpos=amount
        bin=self.GetBinaryData(0,endpos)
        data = [ord(i) for i in bin]
        binary=0
        for ch in data:
            if (ch<8) or (ch>13 and ch<32) or (ch>126):
                binary+=1
        if binary>(endpos/percentage):
            return True
        return False

    def GetSelection2(self):
        """Get the current region, but don't return an empty last line if the
        cursor is at column zero of the last line.
        
        The STC seems to make entire line selections by placing the cursor
        on the left margin of the next line, rather than the end of the
        last highlighted line.  This causes any use of GetLineEndPosition
        to use this line with only the cursor to mean a real part of the
        selection, which is never what I indend, at least.  So, this version of
        GetSelection handles this case.
        """
        start, end = self.GetSelection()
        if start == end:
            return start, end
        if self.GetColumn(end) == 0:
            line = self.LineFromPosition(end - 1)
            newend = self.GetLineEndPosition(line)
            # If the new end is still greater than the start, then we'll assume
            # that this is going to work; otherwise, it wouldn't be possible to
            # select only the newline
            if newend > start:
                return start, newend
        return start, end

    def GetOneLineTarget(self):
        """Create a target encompassing the current line
        
        The STC seems to make entire line selections by placing the cursor
        on the left margin of the next line, rather than the end of the
        last highlighted line.
        """
        start, end = self.GetSelection()
        if start == end:
            line = self.LineFromPosition(start)
            start = self.PositionFromLine(line)
            end = self.GetLineEndPosition(line)
        else:
            start_line = self.LineFromPosition(start)
            end_line = self.LineFromPosition(end)
            # NOTE: The STC seems to make entire line selections by placing the
            # cursor on the left margin of the next line, rather than the end
            # of the last highlighted line.
            if end_line == start_line + 1 and self.GetColumn(end) == 0:
                start = self.PositionFromLine(start_line)
                end = self.GetLineEndPosition(end_line)
            else:
                # There is a selection, and it's more than one line
                return False
        self.SetTargetStart(start)
        self.SetTargetEnd(end)
        return True

    def GetLineRegion(self):
        """Get current region, extending to current line if no region
        selected.

        If there's a region selected, extend it if necessary to
        encompass full lines.  If no region is selected, create one
        from the current line.
        """
        start, end = self.GetSelection()
        if start == end:
            linestart = lineend = self.GetCurrentLine()
        else:
            linestart = self.LineFromPosition(start)
            lineend = self.LineFromPosition(end - 1)

        start -= self.GetColumn(start)
        end = self.GetLineEndPosition(lineend)
        self.SetSelection(start, end)
        return (linestart, lineend)

    def PasteAtColumn(self, paste=None):
        """Paste a rectangular selection at a particular column.
        
        This method inserts a previously cut or copied rectangular selection
        at a column.  If some lines in the STC are too short and end before
        the column, leading spaces are inserted so that the column is pasted
        correctly.
        """
        log.debug("rectangle=%s" % self.SelectionIsRectangle())
        start, end = self.GetSelection()
        log.debug("selection = %d,%d" % (start, end))

        line = self.LineFromPosition(start)
        col = self.GetColumn(start)
        log.debug("line = %d, col=%d" % (line, col))

        if paste is None:
            paste = GetClipboardText()
        self.BeginUndoAction()
        try:
            for insert in paste.splitlines():
                if line >= self.GetLineCount():
                    self.InsertText(self.GetTextLength(), self.getLinesep())
                start = pos = self.PositionFromLine(line)
                last = self.GetLineEndPosition(line)

                # FIXME: doesn't work with tabs
                if (pos + col) > last:
                    # need to insert spaces before the rectangular area
                    num = pos + col - last
                    insert = ' '*num + insert
                    pos = last
                else:
                    pos += col
                log.debug("before: (%d,%d) = '%s'" % (start,last,self.GetTextRange(start,last)))
                log.debug("inserting: '%s' at %d" % (insert, pos))
                self.InsertText(pos, insert)
                log.debug("after: (%d,%d) = '%s'" % (start,last+len(insert),self.GetTextRange(start,last+len(insert))))
                line += 1
        finally:
            self.EndUndoAction()

    def detectLineEndings(self, header=None):
        """Guess which type of line ending is used by the file."""
        def whichLinesep(text):
            # line ending counting function borrowed from PyPE
            crlf_ = text.count('\r\n')
            lf_ = text.count('\n')
            cr_ = text.count('\r')
            mx = max(lf_, cr_)
            if not mx:
                return os.linesep
            elif crlf_ >= mx/2:
                return '\r\n'
            elif lf_ is mx:
                return '\n'
            else:# cr_ is mx:
                return '\r'

        if header is None:
            header = self.GetText()
        linesep = whichLinesep(header)
        mode = self.eol2int[linesep]
        self.SetEOLMode(mode)

    def getNativeEOLMode(self):
        try:
            return self.eol2int[os.linesep]
        except KeyError:
            raise RuntimeError("Unsupported native line separator")

    def isNativeEOLMode(self):
        return self.GetEOLMode() == self.getNativeEOLMode()

    def forceNativeEOLMode(self):
        mode = self.getNativeEOLMode()
        self.ConvertEOLs(mode)

    def ConvertEOLs(self, mode):
        wx.stc.StyledTextCtrl.ConvertEOLs(self, mode)
        self.SetEOLMode(mode)

    def getLinesep(self):
        """Get the current line separator character.

        """
        mode = self.GetEOLMode()
        return self.int2eol[mode]

    def convertStringEOL(self, text):
        """Convert a string to the target EOL format of this STC"""
        target = self.getLinesep()
        if target == '\r':
            text = text.replace('\r\n', '\r').replace('\n', '\r')
        elif target == '\n':
            text = text.replace('\r\n', '\n').replace('\r', '\n')
        else:
            text = text.replace('\r\n', '\n').replace('\r', '\n').replace('\n', '\r\n')
        return text

    # Styling stuff

    def showStyle(self, linenum=None):
        if linenum is None:
            linenum = self.GetCurrentLine()

        linestart = self.PositionFromLine(linenum)

        # actual indention of current line
        ind = self.GetLineIndentation(linenum) # columns
        pos = self.GetLineIndentPosition(linenum) # absolute character position

        # folding says this should be the current indention
        fold = self.GetFoldLevel(linenum)&wx.stc.STC_FOLDLEVELNUMBERMASK - wx.stc.STC_FOLDLEVELBASE

        # get line without indention
        line = self.GetLine(linenum)
        dprint("linenum=%d fold=%d cursor=%d line=%s" % (linenum, fold, self.GetCurrentPos(), repr(line)))
##        for i in range(len(line)):
##            dprint("  pos=%d char=%s style=%d" % (linestart+i, repr(line[i]), self.GetStyleAt(linestart+i) ))

    def showLine(self, line, offset_from_top = 0):
        """Move the cursor to the specified line and show that line at the top
        of the screen.
        
        @param line: line number to display
        
        @keyword offset_from_top: (default 0) optional offset from the top of
        the screen.  The cursor will still be placed at the specified line, but
        the line will be offset C{offset_from_top} number of lines from the top.
        """
        # expand folding if any
        self.EnsureVisible(line)

        # Make line appear at the top of the screen.  ScrollToLine doesn't seem
        # to work when applied right after a GotoLine, but moving to the end
        # of the document and then back does seem to put the line at the top
        # of the screen
        self.GotoLine(self.GetLineCount())

        # If there is an offset from the top of the screen, some extra lines
        # are displayed between the top of the screen and the line with the
        # cursor.
        if offset_from_top > 0:
            self.GotoLine(max(0, line - offset_from_top))
        self.GotoLine(line)
#        dprint("After first GotoLine(%d): FirstVisibleLine = %d" % (line, self.GetFirstVisibleLine()))
#        self.ScrollToLine(line + self.LinesOnScreen() - 3)
#        dprint("After first ScrollToLine(%d): FirstVisibleLine = %d" % (line + self.LinesOnScreen() - 3, self.GetFirstVisibleLine()))
#        self.GotoLine(line)
#        dprint("After second GotoLine(%d): FirstVisibleLine = %d" % (line, self.GetFirstVisibleLine()))
        self.ScrollToColumn(0)

    # --- line indentation stuff

    def FoldAll(self):
        """Fold or unfold all items.
        
        Code borrowed from the wxPython demo
        """
        lineCount = self.GetLineCount()
        expanding = True

        # find out if we are folding or unfolding
        for lineNum in range(lineCount):
            if self.GetFoldLevel(lineNum) & wx.stc.STC_FOLDLEVELHEADERFLAG:
                expanding = not self.GetFoldExpanded(lineNum)
                break

        lineNum = 0

        while lineNum < lineCount:
            level = self.GetFoldLevel(lineNum)
            if level & wx.stc.STC_FOLDLEVELHEADERFLAG and \
               (level & wx.stc.STC_FOLDLEVELNUMBERMASK) == wx.stc.STC_FOLDLEVELBASE:

                if expanding:
                    self.SetFoldExpanded(lineNum, True)
                    lineNum = self.Expand(lineNum, True)
                    lineNum = lineNum - 1
                else:
                    lastChild = self.GetLastChild(lineNum, -1)
                    self.SetFoldExpanded(lineNum, False)

                    if lastChild > lineNum:
                        self.HideLines(lineNum+1, lastChild)

            lineNum = lineNum + 1

    def Expand(self, line, doExpand, force=False, visLevels=0, level=-1):
        """Expand all folds.
        
        Code borrowed from the wxPython demo.
        """
        lastChild = self.GetLastChild(line, level)
        line = line + 1

        while line <= lastChild:
            if force:
                if visLevels > 0:
                    self.ShowLines(line, line)
                else:
                    self.HideLines(line, line)
            else:
                if doExpand:
                    self.ShowLines(line, line)

            if level == -1:
                level = self.GetFoldLevel(line)

            if level & wx.stc.STC_FOLDLEVELHEADERFLAG:
                if force:
                    if visLevels > 1:
                        self.SetFoldExpanded(line, True)
                    else:
                        self.SetFoldExpanded(line, False)

                    line = self.Expand(line, doExpand, force, visLevels-1)

                else:
                    if doExpand and self.GetFoldExpanded(line):
                        line = self.Expand(line, True, force, visLevels-1)
                    else:
                        line = self.Expand(line, False, force, visLevels-1)
            else:
                line = line + 1

        return line

    def GetFoldColumn(self, linenum):
        return self.GetFoldLevel(linenum)&wx.stc.STC_FOLDLEVELNUMBERMASK - wx.stc.STC_FOLDLEVELBASE

    def GetPrevLineIndentation(self, linenum):
        for i in xrange(linenum-1, -1, -1):
            indent = self.GetLineIndentPosition(i)
            last = self.GetLineEndPosition(i)
            if indent<last:
                col = self.GetLineIndentation(i)
                log.debug("line=%d indent=%d (col=%d) last=%d" % (i, indent, col, last))
                return col, i
            log.debug("all blanks: line=%d indent=%d last=%d" % (i, indent, last))
        return 0, -1

    def GetIndentString(self, ind):
        if self.GetUseTabs():
            text = (ind*' ').replace(self.GetTabWidth()*' ', '\t')
        else:
            text = ind*' '
        #dprint("requested: %d, text=%s" % (ind, repr(text)))
        return text

    ##### Utility methods for modifying the contents of the STC
    def addLinePrefixAndSuffix(self, start, end, prefix='', suffix=''):
        """Add a prefix and/or suffix to the line specified by start and end.

        Method to add characters to the start and end of the line. This is
        typically called within a loop that adds comment characters to the
        line.  start and end are assumed to be the endpoints of the
        current line, so no further checking of the line is necessary.

        @param start: first character in line
        @param end: last character in line before line ending
        @param prefix: optional prefix for the line
        @param suffix: optional suffix for the line

        @returns: new position of last character before line ending
        """
        log.debug("commenting %d - %d: '%s'" % (start, end, self.GetTextRange(start,end)))
        slen = len(prefix)
        self.InsertText(start, prefix)
        end += slen

        elen = len(suffix)
        if elen > 0:
            self.InsertText(end, suffix)
            end += elen
        return end + len(self.getLinesep())

    def removeLinePrefixAndSuffix(self, start, end, prefix='', suffix=''):
        """Remove the specified prefix and suffix of the line.

        Method to remove the specified prefix and suffix characters from the
        line specified by start and end.  If the prefix or suffix doesn't match
        the characters in the line, nothing is removed. This is typically
        called within a loop that adds comment characters to the line.  start
        and end are assumed to be the endpoints of the current line, so no
        further checking of the line is necessary.

        @param start: first character in line
        @param end: last character in line before line ending
        @param prefix: optional prefix for the line
        @param suffix: optional suffix for the line

        @returns: new position of last character before line ending
        """
        log.debug("uncommenting %d - %d: '%s'" % (start, end, self.GetTextRange(start,end)))
        slen = len(prefix)
        if self.GetTextRange(start, start+slen) == prefix:
            self.SetSelection(start, start+slen)
            self.ReplaceSelection("")
            end -= slen

        elen = len(suffix)
        if elen > 0:
            if self.GetTextRange(end-elen, end) == suffix:
                self.SetSelection(end-elen, end)
                self.ReplaceSelection("")
                end -= elen
        return end + len(self.getLinesep())

    # Word stuff (for spelling, etc.)
    def getWordFromPosition(self, pos):
        end = self.WordEndPosition(pos, True)
        start = self.WordStartPosition(pos, True)
        word = self.GetTextRange(start, end)
        return (word, start, end)

    def selectBraces(self, pos=None, braces=None):
        """Given a point, find the region contained by the innermost set of braces.
        
        If not specified, all types of braces (e.g.  parens, brackets, curly
        braces) are matched.  Otherwise, only the specified types of braces
        are matched.
        
        @param pos: starting position
        @param braces: string containing brace types
        @return: True if valid region found
        @side-effect: selects region if valid
        """
        if pos is None:
            pos = self.GetCurrentPos()

        if braces is None:
            braces = "([{"
        braces = braces.replace(')', '(').replace(']', '[').replace('}', '{')
        matching = {'(': ')', ')': '(',
                    '[': ']', ']': '[',
                    '{': '}', '}': '{'}

        # use regex to search forward for all brace chars
        pattern = "[\(\[\{\)\]\}]"
        pairs = {u'(': 0, u'[': 0, u'{': 0}
        braceopen = ''
        i = pos
        last = self.GetLength()
        while i < last:
            i = self.FindText(i, last, pattern, wx.stc.STC_FIND_REGEXP)
            if i < 0:
                break
            c = self.GetTextRange(i, i+1)
            s = self.GetStyleAt(i)
            if self.isStyleComment(s) or self.isStyleString(s):
                # Skip matches inside strings
                i += 1
                continue
            if c in u')]}':
                c = matching[c]
                pairs[c] -= 1
                #dprint("-->: closing brace %s at %d: pairs=%s" % (matching[c], i, str(pairs)))
                if pairs[c] < 0:
                    # found an unmatched closing brace.  We're done matching
                    # in this direction, but we have to check if there's a
                    # nesting error.  If all others are zero, we have our
                    # match.  Otherwise, there's some nesting error and we
                    # won't be able to do the search.
                    count = 0
                    for k,v in pairs.iteritems():
                        if k == c:
                            continue
                        count += v
                    if count == 0:
                        # No nesting error, so mark the brace type
                        braceopen = c
                    break
            else:
                # found an opening brace, so now we have to do a search for
                # more brace pairs till we find a closing brace.  We could go
                # many levels of nesting before we find the closing.
                pairs[c] += 1
                #dprint("-->: opening brace %s at %d: pairs=%s" % (c, i, str(pairs)))
            i += 1

        #dprint("brace type = %s at %d" % (braceopen, i))
        last = i

        # Can't use regular expressions searching backward (scintilla
        # limitation), so have to use the slow char-by-char method.
        i = pos
        pairs = {u'(': 0, u'[': 0, u'{': 0}
        while i > 0:
            i -= 1
            c = self.GetTextRange(i, i+1)
            if c in matching:
                if c in u')]}':
                    c = matching[c]
                    pairs[c] += 1
                    #dprint("<--: closing brace %s at %d: pairs=%s" % (matching[c], i, str(pairs)))
                else:
                    pairs[c] -= 1
                    #dprint("<--: opening brace %s at %d: pairs=%s" % (c, i, str(pairs)))
                    if pairs[c] < 0:
                        # found an unmatched opening brace.  Check nesting
                        # and return
                        count = 0
                        for k,v in pairs.iteritems():
                            if k == c:
                                continue
                            count += v
                        if count == 0:
                            # No nesting error, so we have the range
                            #dprint("found match of %s" % braceopen)
                            self.SetSelection(i, last + 1)
                            return True
                        return False
        return False

    def findSameStyle(self, pos=None):
        """Given a point, find the region that has the same style
        
        @param pos: starting position
        @return: tuple of start, end position
        """
        if pos is None:
            pos = self.GetCurrentPos()

        mask = (2**self.GetStyleBits()) - 1
        style = self.GetStyleAt(pos) & mask

        i = pos
        last = self.GetLength()
        while i < last:
            s = self.GetStyleAt(i) & mask
            if s != style:
                break
            i += 1
        last = i

        i = pos
        while i > 0:
            s = self.GetStyleAt(i - 1) & mask
            if s != style:
                break
            i -= 1
        first = i
        return first, last

    ##### Revert hooks
    def getViewPositionData(self):
        return {'top': self.GetFirstVisibleLine(),
                'pos': self.GetCurrentPos(),
                'line': self.GetCurrentLine(),
                }

    def setViewPositionData(self, data):
        if 'top' in data:
            line = min(data['top'], self.GetLineCount() - 1)
            self.ScrollToLine(line)
        if 'line' in data:
            line = min(data['line'], self.GetLineCount() - 1)
            self.showLine(line)
        if 'pos' in data:
            pos = min(data['pos'], self.GetLength() - 1)
            self.GotoPos(pos)

            # Hack to fix #505.  For some reason, the internal scintilla cursor
            # column isn't correct after setting a cursor position manually.
            # Moving the cursor left and then back seems to fix it.
            if pos > 0:
                self.CmdKeyExecute(wx.stc.STC_CMD_CHARLEFT)
                self.CmdKeyExecute(wx.stc.STC_CMD_CHARRIGHT)


class PeppySTC(PeppyBaseSTC):
    """Base class used by major modes that use the STC.
    
    This class contains all the GUI callbacks and mouse bindings on top of
    L{PeppyBaseSTC}
    """

    def __init__(self, parent, refstc=None, copy=None, **kwargs):
        """Initialize the styled text control with peppy extensions.
        
        @param parent: wx.Control used as the parent control
        
        @keyword refstc: (default None) the parent STC, if any, to which
        this document object is linked.  Edits in one document are reflected
        transparently in all the linked documents.
        
        @keyword copy: (default None) STC from which to copy the text into the
        body of this STC
        
        @keyword cmd_key_clear_all: (default True) boolean value to prevent
        the normal call to CmdKeyClearAll.  Normally CmdKeyClearAll is called
        which removes all the default scintilla control keys because the peppy
        actions are used in place.  In rare cases, for example the Editra
        L{StyleEditor}, the scintilla actions are left in place because I've
        not bothered to convert the style editor to use peppy actions.
        """
        if 'cmd_key_clear_all' in kwargs:
            cmd_key_clear_all = kwargs['cmd_key_clear_all']
            del kwargs['cmd_key_clear_all']
        else:
            cmd_key_clear_all = True
        PeppyBaseSTC.__init__(self, parent, refstc=refstc, copy=copy, **kwargs)

        # Only bind events on STCs used as user interface elements.  The
        # reference STC that is not shown doesn't have any way for the user to
        # generate events on it, so event bindings aren't needed.  Plus, this
        # can lead to multiple events and problems deleting the view (bug #770)
        if refstc is not None:
            self.addSTCEventBindings()

        self.modified_callbacks = []

        # Remove all default scintilla keybindings so they will be replaced by
        # peppy actions.
        if cmd_key_clear_all:
            self.CmdKeyClearAll()

        self.debug_dnd=False

    def addSTCEventBindings(self):
        self.Bind(wx.stc.EVT_STC_DO_DROP, self.OnDoDrop)
        self.Bind(wx.stc.EVT_STC_DRAG_OVER, self.OnDragOver)
        self.Bind(wx.stc.EVT_STC_START_DRAG, self.OnStartDrag)
        self.Bind(wx.EVT_MOUSEWHEEL, self.OnMouseWheel)
        self.Bind(wx.EVT_MIDDLE_UP, self.OnMousePaste)
        self.Bind(wx.EVT_LEFT_UP, self.OnSelectionEnd)
        self.Bind(wx.EVT_WINDOW_DESTROY, self.OnDestroy)
        self.Bind(wx.stc.EVT_STC_MODIFIED, self.OnModified)

    def sendEvents(self,evt):
        """
        Send an event to all subordinate STCs
        """
        for otherstc in self.subordinates:
            log.debug("sending event %s to %s" % (evt,otherstc))
            wx.PostEvent(otherstc,evt())

    def addUpdateUIEvent(self, callback):
        self.Bind(wx.stc.EVT_STC_UPDATEUI, callback)

    def addDocumentChangeEvent(self, callback):
        self.Bind(wx.stc.EVT_STC_CHANGE, callback)

    def removeDocumentChangeEvent(self):
        self.Unbind(wx.stc.EVT_STC_CHANGE)

    def OnDestroy(self, evt):
        """
        Event handler for EVT_WINDOW_DESTROY. Preserve the clipboard
        contents can be preserved after the window is destroyed so
        that other apps can still grab it.

        @param evt: event
        """
        wx.TheClipboard.Flush()
        evt.Skip()

    def OnMouseWheel(self, evt):
        """Handle mouse wheel scrolling.
        """
        mouse = wx.GetApp().mouse
        log.debug("Wheel! delta=%s rotation=%s" % (evt.GetWheelDelta(), evt.GetWheelRotation()))
        if mouse.classprefs.mouse_wheel_scroll_style == 'lines':
            num = mouse.classprefs.mouse_wheel_scroll_lines
        else:
            num = self.LinesOnScreen()
            if mouse.classprefs.mouse_wheel_scroll_style == 'half':
                num /= 2
        if evt.GetWheelRotation() > 0:
            # positive means scrolling up, which is a negative number of lines
            # to scroll
            num = -num
        self.LineScroll(0, num)
        #evt.Skip()

    def OnMousePaste(self, evt):
        """Paste the primary selection (on unix) at the mouse cursor location
        
        This currently supports only the primary selection (not the normal
        cut/paste clipboard) from a unix implementation.
        """
        pos = self.PositionFromPoint(wx.Point(evt.GetX(), evt.GetY()))
        #print("x=%d y=%d pos=%d" % (evt.GetX(), evt.GetY(), pos))
        if pos != wx.stc.STC_INVALID_POSITION:
            text = GetClipboardText(True)
            if text:
                text = self.convertStringEOL(text)
                self.BeginUndoAction()
                self.InsertText(pos, text)
                self.GotoPos(pos + len(text))
                self.EndUndoAction()

    def OnSelectionEnd(self, evt):
        """Copy the selected region into the primary selection
        
        This currently supports unix only, because it depends on the primary
        selection of the clipboard.
        """
        start, end = self.GetSelection()
        if start != end:
            text = self.GetTextRange(start, end)
            SetClipboardText(text, True)
        evt.Skip()

    def OnStartDrag(self, evt):
        log.debug("OnStartDrag: %d, %s\n"
                       % (evt.GetDragAllowMove(), evt.GetDragText()))

        if self.debug_dnd and evt.GetPosition() < 250:
            evt.SetDragAllowMove(False)     # you can prevent moving of text (only copy)
            evt.SetDragText("DRAGGED TEXT") # you can change what is dragged
            #evt.SetDragText("")             # or prevent the drag with empty text

    def OnDragOver(self, evt):
        log.debug(
            "OnDragOver: x,y=(%d, %d)  pos: %d  DragResult: %d\n"
            % (evt.GetX(), evt.GetY(), evt.GetPosition(), evt.GetDragResult())
            )

        if self.debug_dnd and evt.GetPosition() < 250:
            evt.SetDragResult(wx.DragNone)   # prevent dropping at the beginning of the buffer

    def OnDoDrop(self, evt):
        log.debug("OnDoDrop: x,y=(%d, %d)  pos: %d  DragResult: %d\n"
                       "\ttext: %s\n"
                       % (evt.GetX(), evt.GetY(), evt.GetPosition(), evt.GetDragResult(),
                          evt.GetDragText()))

        if self.debug_dnd and evt.GetPosition() < 500:
            evt.SetDragText("DROPPED TEXT")  # Can change text if needed
            #evt.SetDragResult(wx.DragNone)  # Can also change the drag operation, but it
                                             # is probably better to do it in OnDragOver so
                                             # there is visual feedback

            #evt.SetPosition(25)             # Can also change position, but I'm not sure why
                                             # you would want to...

    def OnModified(self, evt):
        # NOTE: on really big insertions, evt.GetText can cause a
        # MemoryError on MSW, so I've commented this dprint out.
        #log.debug("(%s) at %d: text=%s" % (self.transModType(evt.GetModificationType()),evt.GetPosition(), repr(evt.GetText())))

        # Since the stc doesn't store the EOL state as an undoable
        # parameter, we have to check for it.
        mod = evt.GetModificationType()
        if mod & (wx.stc.STC_PERFORMED_UNDO | wx.stc.STC_PERFORMED_REDO) and mod & (wx.stc.STC_MOD_INSERTTEXT | wx.stc.STC_MOD_DELETETEXT):
            text = evt.GetText()
            if self.maybe_undo_eolmode is None:
                self.maybe_undo_eolmode = {'total': 0, 'linesep': 0, 'likely': False}
            stats = self.maybe_undo_eolmode
            stats['total'] += 1
            if text == '\n' or text == '\r':
                log.debug("found eol char")
                stats['linesep'] += 1
            if mod & wx.stc.STC_LASTSTEPINUNDOREDO:
                log.debug("eol summary: %s" % stats)
                if stats['linesep'] == stats['total'] and stats['linesep'] >= self.GetLineCount()-1:
                    log.debug("likely that this is a eol change")
                    stats['likely'] = True
        elif mod & wx.stc.STC_MOD_CHANGEFOLD:
            self.OnFoldChanged(evt)
        for cb in self.modified_callbacks:
            cb(evt)
        evt.Skip()

    def addModifyCallback(self, func):
        self.modified_callbacks.append(func)

    def removeModifyCallback(self, func):
        if func in self.modified_callbacks:
            self.modified_callbacks.remove(func)

    def OnFoldChanged(self, evt):
        pass

    def OnUpdateUI(self, evt):
        log.debug("(%s) at %d: text=%s" % (self.transModType(evt.GetModificationType()),evt.GetPosition(), repr(evt.GetText())))
        evt.Skip()

    def transModType(self, modType):
        st = ""
        table = [(wx.stc.STC_MOD_INSERTTEXT, "InsertText"),
                 (wx.stc.STC_MOD_DELETETEXT, "DeleteText"),
                 (wx.stc.STC_MOD_CHANGESTYLE, "ChangeStyle"),
                 (wx.stc.STC_MOD_CHANGEFOLD, "ChangeFold"),
                 (wx.stc.STC_PERFORMED_USER, "UserFlag"),
                 (wx.stc.STC_PERFORMED_UNDO, "Undo"),
                 (wx.stc.STC_PERFORMED_REDO, "Redo"),
                 (wx.stc.STC_LASTSTEPINUNDOREDO, "Last-Undo/Redo"),
                 (wx.stc.STC_MOD_CHANGEMARKER, "ChangeMarker"),
                 (wx.stc.STC_MOD_BEFOREINSERT, "B4-Insert"),
                 (wx.stc.STC_MOD_BEFOREDELETE, "B4-Delete")
                 ]

        for flag,text in table:
            if flag & modType:
                st = st + text + " "

        if not st:
            st = 'UNKNOWN'

        return st

    def showInitialPosition(self, url, options=None):
        log.debug(u"url=%s scheme=%s auth=%s path=%s query=%s fragment=%s" % (url, url.scheme, url.authority, url.path, url.query, url.fragment))
        if url.fragment:
            line = int(url.fragment)
            line -= self.classprefs.line_number_offset
            self.showLine(line)
        if options:
            self.setViewPositionData(options)



#!/usr/bin/env python

"""
A script to build a little example netcdf file for trajectories
"""

import numpy as np
import datetime

from post_gnome import nc_particles

def write_sample_file(filename):
    start_time = datetime.datetime(2010, 11, 3, 12)
    #timesteps = [T.start_time + datetime.timedelta(hour=1) for i in range(10)]
    timesteps = [datetime.datetime(2010, 11, 3, 12, 0),
                 datetime.datetime(2010, 11, 3, 12, 30),
                 datetime.datetime(2010, 11, 3, 13, 0),
                 ]

    trajectory = []

    # first timestep: three particles
    trajectory.append( {'longitude': np.array( [ -88.0,
                                           -88.1,
                                           -88.1,
                                           ], dtype=np.float64 ),
                        'latitude': np.array( [ 28.0, 
                                           28.0,
                                           28.1,
                                           ], dtype=np.float64 ),
                        'depth': np.array( [ 0.0, 
                                             0.1,
                                             0.2,
                                             ], dtype=np.float64 ),
                        'mass': np.array( [ 0.01, 
                                            0.005,
                                            0.007,
                                            ], dtype=np.float64 ),
                        'id': np.array( [ 0, 
                                          1,
                                          2,
                                          ], dtype=np.int32 ),
                        'status_code': np.array( [1,
                                                  2,
                                                  3,
                                                  ], dtype=np.int16,)
                        } )
                      
    # second timestep: four particles
    trajectory.append( {'longitude': np.array( [ -88.0,
                                           -88.1,
                                           -88.1,
                                           -87.9,
                                           ], dtype=np.float64 ),
                        'latitude': np.array( [ 28.0, 
                                           28.0,
                                           28.1,
                                           27.9
                                           ], dtype=np.float64 ),
                        'depth': np.array( [ 0.0, 
                                             0.1,
                                             0.2,
                                             0.1
                                             ], dtype=np.float64 ),
                        'mass': np.array( [ 0.01, 
                                            0.005,
                                            0.007,
                                            0.006,
                                            ], dtype=np.float64 ),
                        'id': np.array( [ 0, 
                                          1,
                                          2,
                                          3,
                                          ], dtype=np.int32 ),
                        'status_code': np.array( [1,
                                                  2,
                                                  3,
                                                  4,
                                                  ], dtype=np.int16,)
                        } )

    # third timestep: two particles
    trajectory.append( {'longitude': np.array( [ -88.0,
                                           -88.1,
                                           ], dtype=np.float64 ),
                        'latitude': np.array( [ 28.0, 
                                           28.0,
                                           ], dtype=np.float64 ),
                        'depth': np.array( [ 0.0, 
                                             0.1,
                                             ], dtype=np.float64 ),
                        'mass': np.array( [ 0.01, 
                                            0.005,
                                            ], dtype=np.float64 ),
                        'id': np.array( [ 1, 
                                          3,
                                          ], dtype=np.int32 ),
                        'status_code': np.array( [2,
                                                  3,
                                                  ], dtype=np.int16,)
                        } )
                      
    writer = nc_particles.Writer(filename,
                                 num_timesteps=len(timesteps),
                                 ref_time=timesteps[0],
                                 )
    for i, time in enumerate(timesteps):
        writer.write_timestep(time, trajectory[i])
    writer.close()

if __name__ == "__main__":
    write_sample_file('test_particles.nc')



                          


#------------------------------------------------------------------------------
# Copyright (c) 2007, Riverbank Computing Limited
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Riverbank Computing Limited
# Description: <Enthought permissions package component>
#------------------------------------------------------------------------------


# Enthought library imports.
from traits.api import Any, Bool, HasTraits, Instance, List

# Local imports.
from package_globals import get_permissions_manager
from permission import Permission


class AdapterBase(HasTraits):
    """This is the base class for object specific adapters."""

    # The object being proxied.
    proxied = Any

    # The list of permissions applied to the proxied object.
    permissions = List(Instance(Permission))

    # Set if the proxied object should be shown when it is enabled.
    show = Bool

    # The desired enabled state of the proxied object, ie. the state set by the
    # application before any permissions were applied.
    _desired_enabled = Bool

    # The desired visible state of the proxied object, ie. the state set by the
    # application before any permissions were applied.
    _desired_visible = Bool

    # The registered adapter types.
    _adapter_types = {}

    @classmethod
    def register_adapter(cls, adapter, *types):
        """Register an adapter type for one or more object types."""

        for t in types:
            cls._adapter_types[t] = adapter

    @classmethod
    def factory(cls, proxied, permissions, show):
        """Return an adapter for the proxied object.  permissions is a list of
        permissions to attach to the object.  show is set if the proxied object
        should be visible when it is disabled.
        """

        # Find a suitable adapter type.
        for object_type, adapter_type in cls._adapter_types.items():
            if isinstance(proxied, object_type):
                break
        else:
            raise TypeError, "no SecureProxy adapter registered for %s" % proxied

        adapter = adapter_type(proxied=proxied, permissions=permissions,
                show=show)

        # Refresh the state of the object when the authentication state of the
        # current user changes.
        get_permissions_manager().user_manager.on_trait_event(adapter._refresh,
                'user_authenticated')

        return adapter

    def adapt(self):
        """Try and adapt the proxied object and return the adapted object if
        successful.  If None is returned a proxy wrapper will be created.
        """

        return None

    def setattr(self, name, value):
        """The default implementation to set an attribute of the proxied
        object.
        """

        setattr(self.proxied, name, value)

    def get_enabled(self):
        """Get the enabled state of the proxied object."""

        raise NotImplementedError

    def set_enabled(self, value):
        """Set the enabled state of the proxied object."""

        raise NotImplementedError

    def update_enabled(self, value):
        """Update the proxied object after a possible new value of the enabled
        attribute.
        """

        if get_permissions_manager().check_permissions(*self.permissions):
            self.set_enabled(value)

            if not self.show:
                self.set_visible(self._desired_visible)
        else:
            self.set_enabled(False)

            if not self.show:
                self.set_visible(False)

        # Save the desired value so that it can be checked if the user becomes
        # authenticated.
        self._desired_enabled = value

    def get_visible(self):
        """Get the visible state of the proxied object."""

        raise NotImplementedError

    def set_visible(self, value):
        """Set the visible state of the proxied object."""

        raise NotImplementedError

    def update_visible(self, value):
        """Update the proxied object after a possible new value of the visible
        attribute.
        """

        if self.show:
            self.set_visible(value)
        elif get_permissions_manager().check_permissions(*self.permissions):
            self.set_visible(value)
        else:
            self.set_visible(False)

        # Save the desired value so that it can be checked if the user becomes
        # authenticated.
        self._desired_visible = value

    def _refresh(self, user):
        """Invoked whenever the current user's authorisation state changes."""

        if not self.show:
            self.update_visible(self._desired_visible)

        self.update_enabled(self._desired_enabled)



#------------------------------------------------------------------------------
#
#  Copyright (c) 2008, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   10/21/2004
#
#------------------------------------------------------------------------------
""" Defines the table editor factory for all traits user interface toolkits.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

from __future__ import absolute_import

from traits.api import (
    Int,
    Float,
    List,
    Instance,
    Str,
    Color,
    Font,
    Any,
    Tuple,
    Dict,
    Enum,
    Trait,
    Bool,
    Callable,
    Range,
    on_trait_change)

from ..editor_factory import EditorFactory

from ..handler import Handler

from ..helper import Orientation

from ..item import Item

from ..table_filter import TableFilter

from ..ui_traits import AView

from ..view import View

from .enum_editor import EnumEditor

#-------------------------------------------------------------------------
#  Constants:
#-------------------------------------------------------------------------

# The filter used to indicate that the user wants to customize the current
# filter
customize_filter = TableFilter(name='Customize...')

#-------------------------------------------------------------------------
#  Trait definitions:
#-------------------------------------------------------------------------

# A trait whose value can be True, False, or a callable function
BoolOrCallable = Trait(False, Bool, Callable)

#-------------------------------------------------------------------------
#  'ToolkitEditorFactory' class:
#-------------------------------------------------------------------------


class ToolkitEditorFactory(EditorFactory):
    """ Editor factory for table editors.
    """

    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    # List of initial table column descriptors
    columns = List(Instance('traitsui.table_column.TableColumn'))

    # List of other table column descriptors (not initially displayed)
    other_columns = List(
        Instance('traitsui.table_column.TableColumn'))

    # The object trait containing the list of column descriptors
    columns_name = Str

    # The desired number of visible rows in the table
    rows = Int

    # The optional extended name of the trait used to specify an external filter
    # for the table data. The value of the trait must either be an instance of
    # TableEditor, a callable that accepts one argument (a table row) and
    # returns True or False to indicate whether the specified object passes the
    # filter or not, or **None** to indicate that no filter is to be applied:
    filter_name = Str

    # Initial filter that should be applied to the table
    filter = Instance('traitsui.table_filter.TableFilter')

    # List of available filters that can be applied to the table
    filters = List(Instance(
        'traitsui.table_filter.TableFilter'))

    # The optional extended trait name of the trait used to notify that the
    # filter has changed and the displayed objects should be updated.
    # It should be an Event.
    update_filter_name = Str

    # Filter object used to allow a user to search the table.
    # NOTE: If left as None, the table will not be searchable.
    search = Instance('traitsui.table_filter.TableFilter')

    # Default context menu to display when any cell is right-clicked
    menu = Instance('traitsui.menu.Menu')

    # Default trait name containg menu
    menu_name = Str

    # Are objects deletable from the table?
    deletable = BoolOrCallable(False)

    # Is the table editable?
    editable = Bool(True)

    # Should the editor become active after the first click
    edit_on_first_click = Bool(True)

    # Can the user reorder the items in the table?
    reorderable = Bool(False)

    # Can the user configure the table columns?
    configurable = Bool(True)

    # Should the cells of the table automatically size to the optimal size?
    auto_size = Bool(True)

    # Mirrors the Qt QSizePolicy.Policy attribute, for horizontal and vertical
    # dimensions.  For these to be useful, set auto_size to False.  If these
    # are None, then the table size policy will not be set in that dimension
    # (for backwards compatibility).
    h_size_policy = Enum(
        None,
        "preferred",
        "fixed",
        "minimum",
        "maximum",
        "expanding",
        "minimum_expanding",
        "ignored")
    v_size_policy = Enum(
        None,
        "preferred",
        "fixed",
        "minimum",
        "maximum",
        "expanding",
        "minimum_expanding",
        "ignored")

    # Should a new row automatically be added to the end of the table to allow
    # the user to create new entries? If True, **row_factory** must be set.
    auto_add = Bool(False)

    # Should the table items be presented in reverse order?
    reverse = Bool(False)

    # The DockWindow graphical theme:
    dock_theme = Any

    # View to use when editing table items.
    # NOTE: If not specified, the table items are not editable in a separate
    # pane of the editor.
    edit_view = AView(' ')

    # The handler to apply to **edit_view**
    edit_view_handler = Instance(Handler)

    # Width to use for the edit view
    edit_view_width = Float(-1.0)

    # Height to use for the edit view
    edit_view_height = Float(-1.0)

    # Layout orientation of the table and its associated editor pane. This
    # attribute applies only if **edit_view** is not ' '.
    orientation = Orientation

    # Is the table sortable by clicking on the column headers?
    sortable = Bool(True)

    # Does sorting affect the model (vs. just the view)?
    sort_model = Bool(False)

    # Should grid lines be shown on the table?
    show_lines = Bool(True)

    # Should the toolbar be displayed? (Note that False will override settings
    # such as 'configurable', etc., and is a quick way to prevent the toolbar
    # from being displayed; but True will not cause a toolbar to appear if one
    # would not otherwise have been displayed)
    show_toolbar = Bool(False)

    # The vertical scroll increment for the table:
    scroll_dy = Range(1, 32)

    # Grid line color
    line_color = Color(0xC4C0A9)

    # Show column labels?
    show_column_labels = Bool(True)

    # Show row labels?
    show_row_labels = Bool(False)

    # Font to use for text in cells
    cell_font = Font

    # Color to use for text in cells
    cell_color = Color('black')

    # Color to use for cell backgrounds
    # The default is the "WindowColor" constant declared in ui.api:
    # we shall set the value in a trait initializer method, in order to avoid
    # circular imports.
    cell_bg_color = Color

    # Color to use for read-only cell backgrounds
    cell_read_only_bg_color = Color(0xF8F7F1)

    # Whether to even-odd alternate the background color.
    alternate_bg_color = Bool(False)

    # Font to use for text in labels
    label_font = Font

    # Color to use for text in labels
    label_color = Color('black')

    # Color to use for label backgrounds
    # The default is the "WindowColor" constant declared in ui.api:
    # we shall set the value in a trait initializer method, in order to avoid
    # circular imports.
    label_bg_color = Color

    # Background color of selected item
    selection_bg_color = Color('light blue', allow_none=True)

    # Color of selected text
    selection_color = Color('black')

    # Height (in pixels) of column labels
    column_label_height = Int(25)

    # Width (in pixels) of row labels
    row_label_width = Int(82)

    # The initial height of each row (<= 0 means use default value):
    row_height = Int(0)

    # The optional extended name of the trait that the indices of the items
    # currently passing the table filter are synced with:
    filtered_indices = Str

    # The selection mode of the table. The meaning of the various values are as
    # follows:
    #
    # row
    #   Entire rows are selected. At most one row can be selected at once.
    #   This is the default.
    # rows
    #   Entire rows are selected. More than one row can be selected at once.
    # column
    #   Entire columns are selected. At most one column can be selected at
    #   once.
    # columns
    #   Entire columns are selected. More than one column can be selected at
    #   once.
    # cell
    #   Single cells are selected. Only one cell can be selected at once.
    # cells
    #   Single cells are selected. More than one cell can be selected at once.
    selection_mode = Enum('row', 'rows', 'column', 'columns', 'cell', 'cells')

    # The optional extended name of the trait that the current selection is
    # synced with:
    selected = Str

    # The optional extended trait name of the trait that the indices of the
    # current selection are synced with:
    selected_indices = Str

    # The optional extended trait name of the trait that should be assigned
    # an ( object, column ) tuple when a table cell is clicked on (Note: If you
    # want to receive repeated clicks on the same cell, make sure the trait is
    # defined as an Event):
    click = Str

    # The optional extended trait name of the trait that should be assigned
    # an ( object, column ) tuple when a table cell is double-clicked on
    # (Note: if you want to receive repeated double-clicks on the same cell,
    # make sure the trait is defined as an Event):
    dclick = Str

    # Called when a table item is selected
    on_select = Any

    # Called when a table item is double clicked
    on_dclick = Any

    # A factory to generate new rows.
    # NOTE: If None, then the user will not be able to add new rows to the
    # table. If not None, then it must be a callable that accepts
    # **row_factory_args** and **row_factory_kw** and returns a new object
    # that can be added to the table.
    row_factory = Any

    # Arguments to pass to the **row_factory** callable when a new row is
    # created
    row_factory_args = Tuple

    # Keyword arguments to pass to the **row_factory** callable when a new row
    # is created
    row_factory_kw = Dict

    # Hooks for replacing parts of the implementation.
    table_view_factory = Callable()
    source_model_factory = Callable()
    model_factory = Callable()

    #-------------------------------------------------------------------------
    #  Traits view definitions:
    #-------------------------------------------------------------------------

    traits_view = View(['{Initial columns}@',
                        Item('columns', resizable=True),
                        '{Other columns}@',
                        Item('other_columns', resizable=True),
                        '|{Columns}<>'],
                       [['deletable{Are items deletable?}', '9',
                         'editable{Are items editable?}', '9',
                         '-[Item Options]>'],
                        ['show_column_labels{Show column labels?}', '9',
                         'configurable{Are columns user configurable?}', '9',
                            'auto_size{Should columns auto size?}',
                            '-[Column Options]>'],
                        ['sortable{Are columns sortable?}',
                         Item('sort_model{Does sorting affect the model?}',
                              enabled_when='sortable'),
                         '-[Sorting Options]>'],
                        [['show_lines{Show grid lines?}',
                          '|>'],
                         ['_', 'line_color{Grid line color}@',
                          '|<>'],
                         '|[Grid Line Options]'],
                        '|{Options}'],
                       [['cell_color{Text color}@',
                         'cell_bg_color{Background color}@',
                         'cell_read_only_bg_color{Read only color}@',
                         '|[Cell Colors]'],
                        ['cell_font',
                         '|[Cell Font]<>'],
                        '|{Cell}'],
                       [['label_color{Text color}@',
                         'label_bg_color{Background color}@',
                         '|[Label Colors]'],
                        ['label_font@',
                         '|[Label Font]<>'],
                        '|{Label}'],
                       [['selection_color{Text color}@',
                         'selection_bg_color{Background color}@',
                         '|[Selection Colors]'],
                        '|{Selection}'],
                       height=0.5)

    #-------------------------------------------------------------------------
    #  'Editor' factory methods:
    #-------------------------------------------------------------------------

    def readonly_editor(self, ui, object, name, description, parent):
        """ Generates an "editor" that is read-only.
        Overridden to set the value of the editable trait to False before
        generating the editor.

        """
        self.editable = False
        return super(ToolkitEditorFactory, self).readonly_editor(
            ui, object, name, description, parent)

    def _cell_bg_color_default(self):
        """ Returns the default value of the cell background color.
        """
        # NOTE: We are initializing the 'cell_bg_color' trait in this method
        # instead of in the trait definition so as to delay importing from
        # ui.api until needed (will lead to circular imports otherwise).
        from ..api import WindowColor
        return WindowColor

    def _label_bg_color_default(self):
        """ Returns the default value of the cell background color.
        """
        # NOTE: We are initializing the 'cell_bg_color' trait in this method
        # instead of in the trait definition so as to delay importing from
        # ui.api until needed (will lead to circular imports otherwise).
        from ..api import WindowColor
        return WindowColor

    #-------------------------------------------------------------------------
    #  Event handlers:
    #-------------------------------------------------------------------------

    @on_trait_change('filters[]')
    def _update_filter_editor(self, object, name, old, new):
        """ Handles the set of filters associated with the editor's factory
            being changed.
        """
        values = {None: '000:No filter'}
        i = 0
        for filter in self.filters:
            if not filter.template:
                i += 1
                values[filter] = '%03d:%s' % (i, filter.name)
        values[customize_filter] = '%03d:%s' % ((i + 1),
                                                customize_filter.name)
        if self._filter_editor is None:
            self._filter_editor = EnumEditor(values=values)
        else:
            self._filter_editor.values = values


# Define the TableEditor class
TableEditor = ToolkitEditorFactory

#-------------------------------------------------------------------------
#  Base class for toolkit-specific editors
#-------------------------------------------------------------------------


class BaseTableEditor(object):
    """ Base class for toolkit-specific editors.
    """

    #-------------------------------------------------------------------------
    #  Interface for toolkit-specific editors:
    #-------------------------------------------------------------------------

    def set_menu_context(self, selection, object, column):
        """Call before creating a context menu for a cell, then set self as the
           controller for the menu.
        """
        self._menu_context = {'selection': selection,
                              'object': object,
                              'column': column,
                              'editor': self,
                              'info': self.ui.info,
                              'handler': self.ui.handler}

    #-------------------------------------------------------------------------
    #  pyface.action 'controller' interface implementation:
    #-------------------------------------------------------------------------

    def add_to_menu(self, menu_item):
        """ Adds a menu item to the menu bar being constructed.
        """
        action = menu_item.item.action
        self.eval_when(action.enabled_when, menu_item, 'enabled')
        self.eval_when(action.checked_when, menu_item, 'checked')

    def add_to_toolbar(self, toolbar_item):
        """ Adds a toolbar item to the toolbar being constructed.
        """
        self.add_to_menu(toolbar_item)

    def can_add_to_menu(self, action):
        """ Returns whether the action should be defined in the user interface.
        """
        if action.defined_when != '':
            if not eval(action.defined_when, globals(),
                        self._menu_context):
                return False

        if action.visible_when != '':
            if not eval(action.visible_when, globals(),
                        self._menu_context):
                return False

        return True

    def can_add_to_toolbar(self, action):
        """ Returns whether the toolbar action should be defined in the user
            interface.
        """
        return self.can_add_to_menu(action)

    def perform(self, action, action_event=None):
        """ Performs the action described by a specified Action object.
        """
        self.ui.do_undoable(self._perform, action)

    def _perform(self, action):
        method_name = action.action
        info = self.ui.info
        handler = self.ui.handler
        context = self._menu_context
        self._menu_context = None
        selection = context['selection']

        if method_name.find('.') >= 0:
            if method_name.find('(') < 0:
                method_name += '()'
            try:
                eval(method_name, globals(), context)
            except:
                # fixme: Should the exception be logged somewhere?
                pass
            return

        method = getattr(handler, method_name, None)
        if method is not None:
            method(info, selection)
            return

        if action.on_perform is not None:
            action.on_perform(selection)
            return

        action.perform(selection)

    #-------------------------------------------------------------------------
    #  Menu support methods:
    #-------------------------------------------------------------------------

    def eval_when(self, condition, object, trait):
        """ Evaluates a condition within a defined context and sets a specified
        object trait based on the result, which is assumed to be a Boolean.
        """
        if condition != '':
            value = bool(eval(condition, globals(), self._menu_context))
            setattr(object, trait, value)

#-------------------------------------------------------------------------
#  Helper class for toolkit-specific editors to implement 'reversed' option:
#-------------------------------------------------------------------------


class ReversedList(object):
    """ A list whose order is the reverse of its input.
    """

    def __init__(self, list):
        self.list = list

    def insert(self, index, value):
        """ Inserts a value at a specified index in the list.
        """
        return self.list.insert(self._index(index - 1), value)

    def index(self, value):
        """ Returns the index of the first occurrence of the specified value in
            the list.
        """
        list = self.list[:]
        list.reverse()

        return list.index(value)

    def __len__(self):
        """ Returns the length of the list.
        """
        return len(self.list)

    def __getitem__(self, index):
        """ Returns the value at a specified index in the list.
        """
        return self.list[self._index(index)]

    def __setslice__(self, i, j, values):
        """ Sets a slice of a list to the contents of a specified sequence.
        """
        return self.list.__setslice__(self._index(i), self._index(j),
                                      values)

    def __delitem__(self, index):
        """ Deletes the item at a specified index.
        """
        return self.list.__delitem__(self._index(index))

    def _index(self, index):
        """ Returns the "reversed" value for a specified index.
        """
        if index < 0:
            return (-1 - index)

        result = (len(self.list) - index - 1)
        if result >= 0:
            return result

        return index



import os
import sys
import time
import argparse
import json

import jsonpickle
from datetime import datetime
import fs
import wx

# Enthought library imports.
from traits.etsconfig.api import ETSConfig
from envisage.ui.tasks.api import TasksApplication
from envisage.ui.tasks.task_window_event import TaskWindowEvent, VetoableTaskWindowEvent
from pyface.api import ImageResource
from pyface.tasks.api import Task, TaskWindowLayout
from traits.api import provides, Bool, Instance, List, Property, Str, Unicode, Event, Dict, Int, Float, Tuple, Any, TraitError
from apptools.preferences.api import Preferences

# Local imports.
from filesystem import init_filesystems
from document import BaseDocument
import documentation
from omnivore.help import get_htmlhelp, MissingDocumentationError
from omnivore.framework.preferences import FrameworkPreferences, \
    FrameworkPreferencesPane
from omnivore.utils.background_http import BackgroundHttpDownloader
import omnivore.utils.wx.error_logger as error_logger

import logging
log = logging.getLogger(__name__)


def _task_window_wx_on_mousewheel(self, event):
    if self.active_task and hasattr(self.active_task, '_wx_on_mousewheel_from_window'):
        log.debug("calling mousewheel in task %s" % self.active_task)
        self.active_task._wx_on_mousewheel_from_window(event)


class FrameworkApplication(TasksApplication):
    """ The sample framework Tasks application.
    """

    #### 'IApplication' interface #############################################

    # The application's globally unique identifier.
    id = 'omnivore.framework.application'

    # The application's user-visible name.
    name = 'Omnivore'

    #### 'TasksApplication' interface #########################################

    # The default window-level layout for the application.
    default_layout = List(TaskWindowLayout)

    # Whether to restore the previous application-level layout when the
    # applicaton is started.
    always_use_default_layout = Property(Bool)

    #### 'FrameworkApplication' interface ####################################

    preferences_helper = Instance(FrameworkPreferences)

    startup_task = Str('omnivore.framework.text_edit_task')

    successfully_loaded_event = Event

    plugin_event = Event

    preferences_changed_event = Event

    plugin_data = {}

    command_line_args = List

    log_dir = Str

    log_file_ext = Str

    cache_dir = Str

    user_data_dir = Str

    next_document_id = Int(0)

    document_class = Any

    documents = List

    last_clipboard_check_time = Float(-1)

    clipboard_check_interval = Float(.75)

    perspectives = Dict

    default_window_size = (800, 600)

    last_window_size = Tuple(default_window_size)

    downloader = Any

    ###########################################################################
    # Private interface.
    ###########################################################################

    #### Trait initializers ###################################################

    def _default_layout_default(self):
        active_task = self.preferences_helper.default_task
        log.debug("active task: -->%s<--" % active_task)
        if not active_task:
            active_task = self.startup_task
        log.debug("active task: -->%s<--" % active_task)
        log.debug("factories: %s" % " ".join([ factory.id for factory in self.task_factories]))
        tasks = [ factory.id for factory in self.task_factories if active_task and active_task == factory.id ]
        log.debug("Default layout: %s" % str(tasks))
        return [ TaskWindowLayout(*tasks,
                                  active_task = active_task,
                                  size = self.last_window_size) ]

    def _preferences_helper_default(self):
        return FrameworkPreferences(preferences = self.preferences)

    def _document_class_default(self):
        return BaseDocument

    #### Trait property getter/setters ########################################

    def _get_always_use_default_layout(self):
        #return self.preferences_helper.always_use_default_layout
        return True

    #### Trait event handlers

    def _application_initialized_fired(self):
        log.debug("STARTING!!!")
        init_filesystems()
        loaded = False
        parser = argparse.ArgumentParser(description="Application parser")
        parser.add_argument("-t", "--task_id", "--task-id", "--edit_with","--edit-with", action="store", default="", help="Use the editing mode specified by this task id for all files listed on the command line")
        parser.add_argument("--show_editors", "--show-editors", action="store_true", default=False, help="List all task ids")
        parser.add_argument("--build_docs", "--build-docs", action="store_true", default=False, help="Build documentation from the menubar")
        options, extra_args = parser.parse_known_args(self.command_line_args)
        if options.show_editors:
            for factory in self.task_factories:
                print("%s %s" % (factory.id, factory.name))
        i = 0
        while i < len(extra_args):
            arg = extra_args[i]
            if arg.startswith("-"):
                if arg == "-d":
                    i += 1
                    error_logger.enable_loggers(extra_args[i])
                else:
                    log.debug("skipping flag %s" % arg)
                i += 1
                continue
            log.debug("processing %s" % arg)
            task_id = self.find_best_task_id(options.task_id)
            self.load_file(arg, None, task_id=task_id)
            loaded = True
            i += 1
        if not loaded:
            factory = self.get_task_factory(self.startup_task)
            url = factory.factory.about_application
            if url:
                self.load_file(url)
        app = wx.GetApp()
        app.tasks_application = self

        if options.build_docs:
            idle = self.on_idle_build_docs
        else:
            idle = self.on_idle
        app.Bind(wx.EVT_IDLE, idle)

    def _application_exiting_fired(self):
        log.debug("CLEANING UP!!!")
        if self.downloader:
            self.downloader = None

    def on_idle(self, evt):
        evt.Skip()
        if not self.active_window:
            return
        editor = self.active_window.active_task.active_editor
        if editor is None:
            return
        t = time.time()
        if t > self.last_clipboard_check_time + self.clipboard_check_interval:
            wx.CallAfter(self.check_clipboard_can_paste, editor)
            self.last_clipboard_check_time = time.time()
        editor.perform_idle()

    def on_idle_build_docs(self, evt):
        evt.Skip()
        if not self.active_window:
            return
        editor = self.active_window.active_task.active_editor
        if editor is None:
            return
        print "Building documentation."
        wx.CallAfter(self.build_docs)

    def build_docs(self):
        import task as frameworktask
        task = frameworktask.FrameworkTask()
        sections = []
        docs = documentation.RSTOnePageDocs("%s %s User's Guide" % (task.about_title, task.about_version), "manual")
        for factory in self.task_factories:
            if "omnivore" not in factory.id or "framework" in factory.id:
                print "Skipping documentation for %s" % factory.id
                continue

            # For testing, uncomment the following block to only process
            # a single task

            # if not factory.id.startswith("omnivore.map_edit"):
            #     print "Skipping documentation for %s" % factory.id
            #     continue

            print "Building documentation for %s (%s)" % (factory.id, factory.name)
            task = self.create_task(factory.id)
            try:
                self.add_task_to_window(self.active_window, task)
                task.new()
            except AttributeError, e:
                print "Error creating documentation for %s: %s" % (factory.id, e)
                continue
            docs.add_task(task)
        docs.create_manual()
        print "Finished creating documentation! Exiting."
        sys.exit()

    def check_clipboard_can_paste(self, editor):
        data_formats = [o.GetFormat() for o in editor.supported_clipboard_data_objects]
        log.debug("Checking clipboard formats %s" % str(data_formats))
        supported = False
        if not wx.TheClipboard.IsOpened():
            try:
                if wx.TheClipboard.Open():
                    for f in data_formats:
                        if wx.TheClipboard.IsSupported(f):
                            log.debug("  found clipboard format: %s" % str(f))
                            supported = True
            finally:
                if wx.TheClipboard.IsOpened():
                    wx.TheClipboard.Close()
            editor.can_paste = supported

    def _window_created_fired(self, event):
        """The toolkit window doesn't exist yet.
        """
        self.init_perspectives()

    def _window_opened_fired(self, event):
        """The toolkit window does exist here.
        """
        log.debug("WINDOW OPENED!!! %s, size=%s" % (event.window.control, str(self.last_window_size)))
        event.window.size = tuple(self.last_window_size)

        # Check to see that there's at least one task.  If a bad application
        # memento (~/.config/Omnivore/tasks/wx/application_memento), the window
        # may be blank in which case we need to add the default task.
        if not event.window.tasks:
            self.create_task_in_window(self.startup_task, event.window)
            log.debug("EMPTY WINDOW OPENED!!! Created task.")

        task = event.window.active_task
        if task.active_editor is None and task.start_new_editor_in_new_window:
            task.new()

        if sys.platform.startswith("win"):
            # monkey patch to include mousewheel handler on the TaskWindow
            import types
            event.window._wx_on_mousewheel = types.MethodType(_task_window_wx_on_mousewheel, event.window)
            event.window.control.Bind(wx.EVT_MOUSEWHEEL, event.window._wx_on_mousewheel)

    #### API

    def guess_document(self, guess):
        service = self.get_service("omnivore.file_type.i_file_recognizer.IFileRecognizerDriver")
        log.debug("SERVICE!!! %s" % service)

        # Attempt to classify the guess using the file recognizer service
        document = service.recognize(guess)
        log.debug("created document %s (mime=%s)" % (document, document.metadata.mime))
        return document

    def load_file(self, uri, active_task=None, task_id="", in_current_window=False, **kwargs):
        from omnivore.utils.file_guess import FileGuess
        # The FileGuess loads the first part of the file and tries to identify it.
        try:
            guess = FileGuess(uri)
        except fs.errors.FSError, e:
            log.error("File load error: %s" % str(e))
            if active_task is not None:
                active_task.window.error(str(e), "File Load Error")
            return

        if len(guess.bytes) == 0:
            if active_task is not None:
                active_task.window.error("Zero length file!\nUnable to determine file type.", "File Load Error")
            return

        # Attempt to classify the guess using the file recognizer service
        document = self.guess_document(guess)
        log.debug("using %s for %s" % (document.__class__.__name__, guess.metadata.uri))
        if document.load_error:
            if self.active_window:
                self.active_window.warning(document.load_error, "Document Load Error")

        # Short circuit: if the file can be edited by the active task, use that!
        if active_task is not None and active_task.can_edit(document):
            log.debug("active task %s can edit %s" % (active_task, document))
            active_task.new(document, **kwargs)
            return

        possibilities = self.get_possible_task_factories(document, task_id)
        if not possibilities:
            log.debug("no editor for %s" % uri)
            return
        best = self.find_best_task_factory(document, possibilities)
        log.debug("best task match: %s" % best.id)

        if active_task is not None:
            # Ask the active task if it's OK to load a different editor
            if not active_task.allow_different_task(guess, best.factory):
                return
            dummy = self.document_class(metadata="application/octet-stream")
            if active_task.can_edit(document) and active_task.ask_attempt_loading_as_octet_stream(guess, best.factory):
                log.debug("Active task %s allows application/octet-stream" % active_task.id)
                active_task.new(document, **kwargs)
                return
            if in_current_window:
                task = self.create_task_in_window(best.id, active_task.window)
                task.new(document, **kwargs)
                return

        # Look for existing task in current windows
        task = self.find_active_task_of_type(best.id)
        if task:
            log.debug("Found task %s in current window" % best.id)
            task.new(document, **kwargs)
            return

        log.debug("Creating task %s in current window" % best.id)
        self.create_task_from_factory_id(document, best.id)

    def get_possible_task_factories(self, document, task_id=""):
        possibilities = []
        for factory in self.task_factories:
            log.debug("checking factory: %s=%s for %s" % (factory.id, factory.name, task_id))
            if task_id:
                if factory.id == task_id or factory.factory.editor_id == task_id:
                    possibilities.append(factory)
            elif hasattr(factory.factory, "can_edit"):
                if factory.factory.can_edit(document):
                    log.debug("  can edit: %s" % document)
                    possibilities.append(factory)
        log.debug("get_possible_task_factories: %s" % str([(p.name, p.id) for p in possibilities]))
        return possibilities

    def find_best_task_factory(self, document, factories):
        scores = []
        for factory in factories:
            log.debug("factory: %s=%s" % (factory.id, factory.name))
            if document.last_task_id == factory.id or document.last_task_id == factory.factory.editor_id:
                # short circuit if document is requesting a specific task
                return factory
            score = factory.factory.get_match_score(document)
            scores.append((score, factory))
        scores.sort()
        log.debug("find_best_task_factory: %s" % str([(s, p.name, p.id) for (s, p) in scores]))
        return scores[-1][1]

    def get_task_factory(self, task_id):
        for factory in self.task_factories:
            if factory.id == task_id or factory.factory.editor_id == task_id:
                return factory
        return None

    def find_best_task_id(self, task_id):
        if task_id:
            for factory in self.task_factories:
                if factory.id == task_id or ".%s." % task_id in factory.id or ".%s" % task_id in factory.id:
                    return factory.id
        return ""  # empty string will result in scanning the file for the best match

    def create_task_from_factory_id(self, guess, factory_id):
        window = self.active_window
        log.debug("  window=%s" % str(window))
        for task in window.tasks:
            if task.id == factory_id:
                break
        else:
            task = self.create_task(factory_id)
        self.add_task_to_window(window, task)
        task.new(guess)
        return task

    def create_task_in_window(self, task_id, window):
        log.debug("creating %s task" % task_id)
        task = self.create_task(task_id)
        self.add_task_to_window(window, task)
        return task

    def add_task_to_window(self, window, task):
        window.add_task(task)
        window.activate_task(task)
        self.restore_perspective(window, task)

    def find_active_task_of_type(self, task_id):
        # Until remove_task bug is fixed, don't create any new windows, just
        # add a new task to the current window unless the task already exists
        w = list(self.windows)
        if not w:
            # OS X might not have any windows open; a menubar is allowed to
            # exist without windows.
            return None
        try:
            i = w.index(self.active_window)
            w[0:0] = [self.active_window]
            w.pop(i)
        except ValueError:
            pass

        for window in w:
            for t in window.tasks:
                if t.id == task_id:
                    log.debug("found non-active task in current window; activating!")
                    window.activate_task(t)
                    return t
        if window:
            task = self.create_task_in_window(task_id, window)
            return task
#        # Check active window first, then other windows
#        w = list(self.windows)
#        try:
#            i = w.index(self.active_window)
#            w[0:0] = [self.active_window]
#            w.pop(i)
#        except ValueError:
#            pass
#
#        for window in w:
#            log.debug("window: %s" % window)
#            log.debug("  active task: %s" % window.active_task)
#            if window.active_task.id == task_id:
#                log.debug("  found active task")
#                return window.active_task
#        log.debug("  no active task matches %s" % task_id)
#        for window in w:
#            task = window.active_task
#            if task is None:
#                continue
#            # if no editors in the task, replace the task with the new task
#            log.debug("  window %s: %d" % (window, len(task.editor_area.editors)))
#            if len(task.editor_area.editors) == 0:
#                log.debug("  replacing unused task!")
#                # The bugs in remove_task seem to have been fixed so that the
#                # subsequent adding of a new task does seem to work now.  But
#                # I'm leaving in the workaround for now of simply closing the
#                # active window, forcing the new task to open in a new window.
#                if True:
#                    log.debug("removing task %s" % task)
#                    print window
#                    #window.remove_task(task)
#                    task = self.create_task_in_window(task_id, window)
#                    return task
#                else:
#                    window.close()
#                    return None

    def find_or_create_task_of_type(self, task_id):
        task = self.find_active_task_of_type(task_id)
        if not task:
            log.debug("task %s not found in active windows; creating new window" % task_id)
            window = self.create_window()
            task = self.create_task_in_window(task_id, window)
            window.open()
        return task

    # Override the default window closing event handlers only on Mac because
    # Mac allows the application to remain open while no windows are open
    if sys.platform == "darwin":
        def _on_window_closing(self, window, trait_name, event):
            # Event notification.
            self.window_closing = window_event = VetoableTaskWindowEvent(
                window=window)

            if window_event.veto:
                event.veto = True
            else:
                # Store the layout of the window.
                window_layout = window.get_window_layout()
                self._state.push_window_layout(window_layout)

        def _on_window_closed(self, window, trait_name, event):
            self.windows.remove(window)

            # Event notification.
            self.window_closed = TaskWindowEvent(window=window)

            # Was this the last window?
            if len(self.windows) == 0 and self._explicit_exit:
                self.stop()

    def _initialize_application_home(self):
        """Override the envisage.application method to force the use of standard
        config directory location instead of ~/.enthought 
        """

        from omnivore.third_party.appdirs import user_config_dir, user_log_dir, user_cache_dir, user_data_dir
        dirname = user_config_dir(self.name)
        ETSConfig.application_home = dirname

        # Make sure it exists!
        if not os.path.exists(ETSConfig.application_home):
            os.makedirs(ETSConfig.application_home)

        dirname = user_log_dir(self.name)
        if not os.path.exists(dirname):
            os.makedirs(dirname)
        self.log_dir = dirname

        self.log_file_ext = "-%s" % datetime.now().strftime("%Y%m%d-%H%M%S")

        dirname = user_cache_dir(self.name)
        if not os.path.exists(dirname):
            os.makedirs(dirname)
        self.cache_dir = dirname

        dirname = user_data_dir(self.name)
        if not os.path.exists(dirname):
            os.makedirs(dirname)
        self.user_data_dir = dirname

        # Prevent py2exe from creating a dialog box on exit saying that there
        # are error messages.  It thinks that anything written to stderr is an
        # error, and the python logging module redirects everything to stderr.
        # Instead, redirect stderr to a log file in the user log directory
        frozen = getattr(sys, 'frozen', False)
        if frozen in ('dll', 'windows_exe', 'console_exe'):
            # redirect py2exe stderr/stdout to log file
            log = self.get_log_file_name("py2exe")
            oldlog = sys.stdout
            sys.stdout = open(log, 'w')
            if hasattr(oldlog, "saved_text"):
                sys.stdout.write("".join(oldlog.saved_text))
            sys.stderr = sys.stdout

            # The logging module won't redirect to the new stderr without help
            handler = logging.StreamHandler(sys.stderr)
            logger = logging.getLogger('')
            logger.addHandler(handler)
        else:
            log = self.get_log_file_name("log")
            handler = logging.FileHandler(log)
            formatter = logging.Formatter("%(levelname)s:%(name)s:%(msg)s")
            handler.setFormatter(formatter)
            logger = logging.getLogger('')
            logger.addHandler(handler)

    #### Convenience methods

    def add_document(self, document):
        """Add document to the application list of open documents
        
        FIXME: check for duplicates?
        """
        existing = self.get_document(document.document_id)
        if existing:
            return existing

        document.document_id = self.next_document_id
        self.next_document_id += 1
        self.documents.append(document)
        return document

    def get_document(self, document_id):
        """Find an existing document given the document_id
        """
        for doc in self.documents:
            if doc.document_id == document_id:
                return doc
        return None

    def get_plugin_data(self, plugin_id):
        return self.plugin_data[plugin_id]

    def get_preferences(self, helper_object, debug=True):
        """Get preferences for a particular PreferenceHelper object.
        
        Handle mistakes in preference files by using the default value for any
        bad preference values.
        """

        # Give the helper application a class attribute to the application
        helper_object.application = self

        try:
            helper = helper_object(preferences=self.preferences)
        except TraitError:
            # Create an empty preference object and helper so we can try
            # preferences one-by-one to see which are bad
            empty = Preferences()
            helper = helper_object(preferences=empty)
            if debug:
                log.debug("Application preferences before determining error:")
                self.preferences.dump()
            for t in helper.trait_names():
                if helper._is_preference_trait(t):
                    pref_name = "%s.%s" % (helper.preferences_path, t)
                    text_value = self.preferences.get(pref_name)
                    if text_value is None:
                        # None means the preference isn't specified, which
                        # isn't an error.
                        continue
                    try:
                        empty.set(pref_name, self.preferences.get(pref_name))
                    except:
                        log.error("Invalid preference for %s: %s. Using default value %s" % (pref_name, self.preferences.get(pref_name), getattr(helper, t)))
                        self.preferences.remove(pref_name)
                        # Also remove from default scope
                        self.preferences.remove("default/%s" % pref_name)
            if debug:
                log.debug("Application preferences after removing bad preferences:")
                self.preferences.dump()
            helper = helper_object(preferences=self.preferences)
        return helper

    def get_log_file_name(self, log_file_name_base, ext=""):
        filename = log_file_name_base + self.log_file_ext
        if ext:
            if not ext.startswith("."):
                filename += "."
            filename += ext
        else:
            filename += ".log"
        filename = os.path.join(self.log_dir, filename)
        return filename

    def save_log(self, text, log_file_name_base, ext=""):
        filename = self.get_log_file_name(log_file_name_base, ext)

        try:
            with open(filename, "wb") as fh:
                fh.write(text)
        except IOError:
            log.error("Failed writing %s to %s" % (log_file_name_base, filename))

    def get_config_dir_filename(self, subdir, json_name):
        config_dir = ETSConfig.application_home
        dirname = os.path.join(config_dir, subdir)
        if not os.path.exists(dirname):
            os.makedirs(dirname)
        return os.path.join(dirname, json_name)

    def get_json_data(self, json_name, default_on_error=None):
        try:
            file_path = self.get_config_dir_filename("json", json_name)
            with open(file_path, "r") as fh:
                raw = fh.read()
            json_data = jsonpickle.decode(raw)
            try:
                # new format is a list with a format identifier as the first entry
                if json_data[0] == "format=v2":
                    decoded = json_data[1]
            except KeyError:
                # deprecated format was a dictionary, and was creating an extra
                # layer of indirection by encoding the jsonpickle as another string
                json_data = json.loads(raw)
                encoded = json_data[json_name]
                decoded = jsonpickle.decode(encoded)
            return decoded
        except IOError:
            # file not found
            return default_on_error
        except ValueError:
            # bad JSON format
            log.error("Bad JSON format in preferences file: %s" % json_name)
            return default_on_error

    def save_json_data(self, json_name, data):
        file_path = self.get_config_dir_filename("json", json_name)
        json_data = ["format=v2", data]
        encoded = jsonpickle.encode(json_data)
        with open(file_path, "w") as fh:
            fh.write(encoded)

    def get_bson_data(self, bson_name):
        import bson

        file_path = self.get_config_dir_filename("bson", bson_name)
        with open(file_path, "r") as fh:
            raw = fh.read()
        if len(raw) > 0:
            bson_data = bson.loads(raw)
            data = bson_data[bson_name]
        else:
            raise IOError("Blank BSON data")
        return data

    def save_bson_data(self, bson_name, data):
        import bson

        file_path = self.get_config_dir_filename("bson", bson_name)
        bson_data = {bson_name: data}
        raw = bson.dumps(bson_data)
        with open(file_path, "w") as fh:
            fh.write(raw)

    # class attributes

    perspectives_loaded = False

    def init_perspectives(self):
        if self.perspectives_loaded:
            return
        data = self.get_json_data("perspectives", default_on_error={})
        self.perspectives = data.get("perspectives", {})
        self.last_window_size = tuple(data.get("window_size", self.default_window_size))
        log.debug("init_perspectives: size=%s" % str(self.last_window_size))

    def remember_perspectives(self, window):
        layout = window.get_layout()
        if layout is not None:
            p = layout.perspective
            # make sure central pane exists, otherwise will restore without
            # central pane visible
            if "name=Central;" in p:
                log.debug("remember_perspective: %s, size=%s" % (layout.id, str(window.size)))
                self.perspectives[layout.id] = p
                self.last_window_size = window.size
                data = {"perspectives": dict(self.perspectives),
                        "window_size": list(self.last_window_size),
                        }
                self.save_json_data("perspectives", data)

    def restore_perspective(self, window, task):
        # get layout object that can be used to restore perspective since we
        # only serialize the perspective string and not the whole layout object
        layout = window.get_layout()
        if task.id in self.perspectives:
            layout.perspective = self.perspectives[task.id]
            window.set_layout(layout)
        task.restore_toolbars(window)

    def get_downloader(self):
        if self.downloader is None:
            self.downloader = BackgroundHttpDownloader()
        return self.downloader

    help_frame = None

    def show_help(self, section=None):
        from wx.html import HtmlHelpController
        if self.help_frame is None:
            filename = self.get_config_dir_filename(".", "htmlhelp.cfg")
            cfg = wx.FileConfig(localFilename=filename, style=wx.CONFIG_USE_LOCAL_FILE)
            # NOTE: using a FileConfig directly in the HtmlHelpController by
            # self.help_frame.UseConfig(cfg) crashes when closing the help
            # window
            wx.ConfigBase.Set(cfg)
            self.help_frame = HtmlHelpController()

        try:
            filename = get_htmlhelp("userguide.hhp")
        except MissingDocumentationError, e:
            self.active_window.warning(str(e), "Help Files Not Found")
            return

        self.help_frame.AddBook(filename)
        # plugins = wx.GetApp().plugin_manager.getActivePluginObjects()
        # for plugin in plugins:
        #     for book in plugin.getHelpBooks():
        #         self.help_frame.AddBook(book)
        if section:
            self.help_frame.Display(section)

            # Make sure it actually displayed something, otherwise show
            # the work-in-progress page
            data = self.help_frame.GetHelpWindow().GetData()
            filename = data.FindPageByName(section)
            if not filename:
                # FIXME: change this to something in omnivore
                self.help_frame.Display("work-in-progress.html")
        else:
            self.help_frame.DisplayContents()



""" Context adapter for trait lists. """


# Enthought library imports.
from traits.api import Any, List, Property, Str

# Local imports.
from list_context_adapter import ListContextAdapter


class TraitListContextAdapter(ListContextAdapter):
    """ Context adapter for trait lists. """

    #### 'Context' interface ##################################################

    # The name of the context within its own namespace.
    namespace_name = Property(Str)

    #### 'ContextAdapter' interface ###########################################

    # The object that we are adapting.
    adaptee = Any

    #### 'ListContextAdapter' interface #######################################

    # The list that we are adapting.
    collection = Property(List)

    #### 'TraitListContextAdapter' interface ##################################

    # The name of the object's trait that provides the list.
    trait_name = Str

    ###########################################################################
    # 'Context' interface.
    ###########################################################################

    def _get_namespace_name(self):
        """ Returns the name of the context within its own namespace. """

        return self.context.namespace_name + '/' + self.trait_name

    ###########################################################################
    # Protected 'ListContext' interface.
    ###########################################################################

    #### 'Properties' #########################################################

    def _get_collection(self):
        """ Returns the collection that we are adapting. """

        components = self.trait_name.split('.')
        if len(components) == 1:
            collection = getattr(self.adaptee, self.trait_name)

        else:
            # Find the object that contains the trait.
            obj = self.adaptee
            for component in components[:-1]:
                obj = getattr(obj, component)

            collection = getattr(obj, components[-1])

        return collection

#### EOF ######################################################################



#------------------------------------------------------------------------------
# Copyright (c) 2005-2009 Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#------------------------------------------------------------------------------
""" Functions to determine resource locations from the call stack.

This type of resource location is normally requested from the constructor for
an object whose resources are relative to the module constructing the object.

"""

import sys

from traits.trait_base import get_resource_path


def resource_module(level = 2):
    """Return a module reference calculated from the caller's stack.

    Note that what we want is the reference to the package containing the
    module in the stack.  This is because we need a directory to search
    for our default resource sub-dirs as children.

    """
    module_name = sys._getframe(level).f_globals.get('__name__', '__main__')
    if '.' in module_name:
        module_name = '.'.join(module_name.split('.')[:-1])
    module = sys.modules.get(module_name)

    return module


def resource_path(level = 2):
    """Return a resource path calculated from the caller's stack.

    """
    return get_resource_path(level + 1)

#### EOF ######################################################################



#------------------------------------------------------------------------------
# Copyright (c) 2008, Riverbank Computing Limited
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Riverbank Computing Limited
# Description: <Enthought application scripting package component>
#------------------------------------------------------------------------------

from scriptable_type import create_scriptable_type, make_object_scriptable
from i_bind_event import IBindEvent
from i_script_manager import IScriptManager
from package_globals import get_script_manager, set_script_manager
from script_manager import ScriptManager
from scriptable import scriptable, Scriptable



import numpy as np

from pyface.tasks.topological_sort import topological_sort

import logging
log = logging.getLogger(__name__)


def find_wildcard_matches(item_map, pattern):
    if pattern.endswith("*"):
        pattern = pattern[:-1]
    if pattern:
        for id, item in item_map.iteritems():
            if id.startswith(pattern):
                yield item


def before_after_wildcard_sort(items):
    """ Sort a sequence of items with 'before', 'after', and 'id' attributes.
        
    The sort is topological. If an item does not specify a 'before' or 'after',
    it is placed after the preceding item.
    
    Simple wildcards are allowed in the 'before' or 'after' attributes.  The
    asterisk must be the last character in the string, so the form "text*"
    will match any id that starts with "text".

    If a cycle is found in the dependencies, a warning is logged and the order
    of the items is undefined.
    """
    # Handle a degenerate case for which the logic below will fail (because
    # prev_item will not be set).
    if len(items) < 2:
        return items

    # Build a set of pairs representing the graph.
    item_map = dict((item.id, item) for item in items if item.id)
    pairs = []
    prev_item = None
    for item in items:
        # Attempt to use 'before' and 'after' to make pairs.
        new_pairs = []
        if hasattr(item, 'before') and item.before:
            if item.before.endswith("*"):
                for child in find_wildcard_matches(item_map, item.before):
#                    print "%s should be before %s" % (item.id, child.id)
                    new_pairs.append((item, child))
            else:
                child = item_map.get(item.before)
                if child:
                    new_pairs.append((item, child))
        if hasattr(item, 'after') and item.after:
            if item.after.endswith("*"):
                for parent in find_wildcard_matches(item_map, item.after):
#                    print "%s should be after %s" % (item.id, parent.id)
                    new_pairs.append((parent, item))
            else:
                parent = item_map.get(item.after)
                if parent:
                    new_pairs.append((parent, item))

        # If we have any pairs, use them. Otherwise, use the previous unmatched
        # item as a parent, if possible.
        if new_pairs:
            pairs.extend(new_pairs)
        else:
            if prev_item:
                pairs.append((prev_item, item))
            prev_item = item

    # Now perform the actual sort.
    result, has_cycle = topological_sort(pairs)
    if has_cycle:
        log.error('Indeterminate result; cycle in before/after sort for items %r', items)
    return result


def collapse_overlapping_ranges(ranges):
    """ Collapse the list of (possibly overlapping) selected ranges into
    a monotonically increasing set of non-overlapping ranges
    """
    opt = []
    start, end = None, None
    for next_start, next_end in sorted(ranges):
        if next_start > next_end:
            next_start, next_end = next_end, next_start
        if start is None:
            start, end = next_start, next_end
        else:
            if next_start > end:
                opt.append((start, end))
                start, end = next_start, next_end
            elif next_end > end:
                end = next_end
    if start is not None:
        opt.append((start, end))
    return opt


def ranges_to_indexes(ranges):
    if len(ranges) == 0:
        return np.zeros([0], dtype=np.uint32)
    return np.hstack((np.arange(r[0], r[1], dtype=np.uint32) for r in ranges))


def indexes_to_ranges(indexes):
    groups = np.split(indexes, np.where(np.diff(indexes) != 1)[0] + 1)
    ranges = []
    for group in groups:
        if np.alen(group) > 0:
            ranges.append((int(group[0]), int(group[-1]) + 1))
    return ranges


def invert_ranges(ranges, last):
    """ Invert the list of (possibly overlapping) selected ranges into a
    monotonically increasing set of non-overlapping ranges that represents the
    opposite of the listed ranges
    """
    # get a monotonically increasing list
    ranges = collapse_overlapping_ranges(ranges)
    inverted = []

    first = 0
    for start, end in ranges:
        if start > first:
            inverted.append((first, start))
            first = end
        else:
            first = end
    if first < last:
        inverted.append((first, last))
    return inverted


def invert_rects(rects, numr, numc):
    # Purely heuristic approach.  An algorithmic approach might be based on:
    # http://stackoverflow.com/questions/30818645 but this one breaks up the
    # entire space into a set of rectangles with boundaries at every possible
    # row/col boundary of all the contained rectangles.

    # Get list of all referenced rows and cols
    rows = sorted(list(set([r[0] for rect in rects for r in rect]).union(set([0, numr]))))
    cols = sorted(list(set([r[1] for rect in rects for r in rect]).union(set([0, numc]))))

    # create inside/outside flags for each row/col in the entire grid
    inside = np.zeros((numr, numc), dtype=np.bool_)
    for [(r1, c1), (r2, c2)] in rects:
        inside[r1:r2, c1:c2] = True

    # create lo/hi values for each subdivision of the grid
    rowpairs = [(rows[i], rows[i+1]) for i in range(len(rows) - 1)]
    colpairs = [(cols[i], cols[i+1]) for i in range(len(cols) - 1)]

    # create a rectangle at each intersection point and if it's outside the
    # original set of rectangles, add it to the list
    outside = []
    for r1, r2 in rowpairs:
        for c1, c2 in colpairs:
            if not inside[r1, c1]:
                outside.append([(r1, c1), (r2, c2)])

    # optimization: merge neighboring rectangles that share common edges

    return outside



#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought naming package component>
#------------------------------------------------------------------------------
""" The node type for contexts in a naming system. """


# Enthought library imports.
from apptools.naming.api import Context
from pyface.tree.api import NodeType

# Local imports.
from context_monitor import ContextMonitor


class ContextNodeType(NodeType):
    """ The node type for contexts in a naming system. """

    ###########################################################################
    # 'NodeType' interface.
    ###########################################################################

    # 'node' in this case is a 'Binding' instance whose 'obj' trait is a
    # 'Context' instance.
    def is_type_for(self, node):
        """ Returns True if this node type recognizes a node. """

        return isinstance(node.obj, Context)

    def allows_children(self, node):
        """ Does the node allow children (ie. a folder vs a file). """

        return True

    def has_children(self, node):
        """ Returns True if a node has children, otherwise False. """

        return len(node.obj.list_names('')) > 0

    def get_children(self, node):
        """ Returns the children of a node. """

        return node.obj.list_bindings('')

    def get_drag_value(self, node):
        """ Get the value that is dragged for a node. """

        return node.obj

    def is_editable(self, node):
        """ Returns True if the node is editable, otherwise False. """

        # You can't rename the root context!
        return node.context is not None

    def get_text(self, node):
        """ Returns the label text for a node. """

        return node.name

    def can_set_text(self, node, text):
        """ Returns True if the node's label can be set. """

        # The parent context will NOT be None here (see 'is_editable').
        parent = node.context

        return len(text.strip()) > 0 and text not in parent.list_names('')

    def set_text(self, node, text):
        """ Sets the label text for a node. """

        print 'Setting text on', node.name, node.obj
        print 'Context details', node.obj.name, node.obj.path

        # Do the rename in the naming system.
        node.context.rename(node.name, text)

        # Update the binding.
        node.name = text

        return

    def get_monitor(self, node):
        """ Returns a monitor that detects changes to a node. """

        return ContextMonitor(node=node)

##### EOF #####################################################################



#------------------------------------------------------------------------------
#
#  Copyright (c) 2012, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Author: Pietro Berkes
#  Date:   Feb 2012
#
#------------------------------------------------------------------------------

"""
Test the creation and layout of labels.
"""

from traits.has_traits import HasTraits
from traits.trait_types import Bool, Str
from traitsui.view import View
from traitsui.item import Item
from traitsui.group import VGroup, HGroup

from traitsui.tests._tools import *


_DIALOG_WIDTH = 500


class ShowRightLabelsDialog(HasTraits):
    """ Dialog with labels on the left/right to test the label text.
    """

    bool_item = Bool(True)

    traits_view = View(
        VGroup(
            VGroup(
                Item('bool_item'),
                show_left=False
            ),
            VGroup(
                Item('bool_item'),
                show_left=True
            ),
        ),
    )


class HResizeTestDialog(HasTraits):
    """ Dialog with checkbox and text elements and labels on the right.
    We test the separation between element and label in HGroups.
    """

    bool_item = Bool(True)
    txt_item = Str

    traits_view = View(
        VGroup(
            HGroup(
                Item('bool_item', springy=True),
                show_left=False
            ),
            VGroup(
                Item('txt_item', resizable=True),
                show_left=False
            ),
        ),
        width=_DIALOG_WIDTH,
        height=100,
        resizable=True
    )


class VResizeTestDialog(HasTraits):
    """ Dialog with checkbox and text elements and labels on the right.
    We test the separation between element and label in VGroups.
    """

    bool_item = Bool(True)
    txt_item = Str

    traits_view = View(
        VGroup(
            VGroup(
                Item('bool_item', resizable=True),
                show_left=False
            ),
            VGroup(
                Item('txt_item', resizable=True),
                show_left=False
            ),
        ),
        width=_DIALOG_WIDTH,
        height=100,
        resizable=True
    )


class NoLabelResizeTestDialog(HasTraits):
    """ Test the combination show_label=False, show_left=False.
    """

    bool_item = Bool(True)

    traits_view = View(
        VGroup(
            Item('bool_item', resizable=True, show_label=False),
            show_left=False
        ),
        resizable=True
    )


@skip_if_not_qt4
def test_qt_show_labels_right_without_colon():
    # Behavior: traitsui should not append a colon ':' to labels
    # that are shown to the *right* of the corresponding elements

    from pyface import qt

    with store_exceptions_on_all_threads():
        dialog = ShowRightLabelsDialog()
        ui = dialog.edit_traits()

        # get reference to label objects
        labels = ui.control.findChildren(qt.QtGui.QLabel)

        # the first is shown to the right, so no colon
        nose.tools.assert_false(labels[0].text().endswith(':'))

        # the second is shown to the right, it should have a colon
        nose.tools.assert_true(labels[1].text().endswith(':'))


def _test_qt_labels_right_resizing(dialog_class):
    # Bug: In the Qt backend, resizing a checkbox element with a label on the
    # right resizes the checkbox, even though it cannot be.
    # The final effect is that the label remains attached to the right margin,
    # with a big gap between it and the checkbox. In this case, the label
    # should be made resizable instead.
    # On the other hand, a text element should keep the current behavior and
    # resize.

    from pyface import qt

    with store_exceptions_on_all_threads():
        dialog = dialog_class()
        ui = dialog.edit_traits()

        # all labels
        labels = ui.control.findChildren(qt.QtGui.QLabel)

        # the checkbox and its label should be close to one another; the
        # size of the checkbox should be small
        checkbox_label = labels[0]
        checkbox = ui.control.findChild(qt.QtGui.QCheckBox)

        # horizontal space between checkbox and label should be small
        h_space = checkbox_label.x() - checkbox.x()
        nose.tools.assert_less(h_space, 100)
        # and the checkbox size should also be small
        nose.tools.assert_less(checkbox.width(), 100)

        # the text item and its label should be close to one another; the
        # size of the text item should be large
        text_label = labels[0]
        text = ui.control.findChild(qt.QtGui.QLineEdit)

        # horizontal space between text and label should be small
        h_space = text_label.x() - text.x()
        nose.tools.assert_less(h_space, 100)
        # and the text item size should be large
        nose.tools.assert_greater(text.width(), _DIALOG_WIDTH - 200)

        # the size of the window should still be 500
        nose.tools.assert_equal(ui.control.width(), _DIALOG_WIDTH)


@skip_if_not_qt4
def test_qt_labels_right_resizing_vertical():
    _test_qt_labels_right_resizing(VResizeTestDialog)


@skip_if_not_qt4
def test_qt_labels_right_resizing_horizontal():
    _test_qt_labels_right_resizing(HResizeTestDialog)


@skip_if_not_qt4
def test_qt_no_labels_on_the_right_bug():
    # Bug: If one set show_left=False, show_label=False on a non-resizable
    # item like a checkbox, the Qt backend tried to set the label's size
    # policy and failed because label=None.

    with store_exceptions_on_all_threads():
        dialog = NoLabelResizeTestDialog()
        ui = dialog.edit_traits()


if __name__ == "__main__":
    # Execute from command line for manual testing
    vw = HResizeTestDialog()
    vw.configure_traits()



# peppy Copyright (c) 2006-2010 Rob McMullen
# HTML conversion subroutines Copyright (c) 2010 Christopher Barker
# This file is licenced under the same terms as Python itself
"""Utility functions for operating on text

These text utilities have no dependencies on any other part of peppy, and
therefore may be used independently of peppy.
"""
import re

import logging
log = logging.getLogger(__name__)


def piglatin(text):
    """Translate string to pig latin.
    
    Simple pig latin translator that properly capitalizes the resulting string,
    and skips over any leading or trailing non-alphabetic characters.
    """
    words = []
    for w in text.split():
        # Find non alphabetic chars at the start and skip them
        i=0
        while not w[i].isalpha():
            i += 1
        start = w[0:i]
        w = w[i:]

        if w[0] in 'aeiouAEIOU':
            prefix = w
            suffix = 'way'
        else:
            if w[0].isupper():
                prefix = w[1:].capitalize()
                suffix = w[0].lower() + 'ay'
            else:
                prefix = w[1:]
                suffix = w[0].lower() + 'ay'

        # Move any trailing non-alphabetic characters to the end
        i = len(prefix) - 1
        while i >= 0 and not prefix[i].isalpha():
            i -= 1
        end = prefix[i + 1:]
        prefix = prefix[0:i + 1]

        word = start + prefix + suffix + end
        #print "preprefix=%s, prefix=%s, suffix=%s, word=%s" % (preprefix, prefix, suffix, word)
        words.append(word)
    return u' '.join(words)


def getMagicComments(bytes, headersize=1024):
    """Given a byte string, get the first two lines.
    
    "Magic comments" appear in the first two lines of the file, and can
    indicate the encoding of the file or the major mode in which the file
    should be interpreted.
    """
    numbytes = len(bytes)
    if headersize > numbytes:
        headersize = numbytes
    header = bytes[0:headersize]
    lines = header.splitlines()
    return lines[0:2]


def detectBOM(bytes):
    """Search for unicode Byte Order Marks (BOM)
    """
    boms = {
        'utf-8': '\xef\xbb\xbf',
        'utf-16-le': '\xff\xfe',
        'utf-16-be': '\xfe\xff',
        'utf-32-le': '\xff\xfe\x00\x00',
        'utf-32-be': '\x00\x00\xfe\xff',
        }
    # FIXME: utf-32 is not available in python 2.5, only 2.6 and later.

    for encoding, bom in boms.iteritems():
        if bytes.startswith(bom):
            return encoding, bom
    return None, None


def detectEncoding(bytes):
    """Search for "magic comments" that specify the encoding
    
    @returns tuple containing (encoding name, BOM) where both may be None.  If
    the Byte Order Mark is not None, the bytes should be stripped off before
    decoding using the encoding name.
    """
    if isinstance(bytes, str):
        # Only check byte order marks when the input is a raw string.  If the
        # input is unicode, it can't have a valid byte order mark.
        encoding, bom = detectBOM(bytes)
        if encoding:
            return encoding, bom
    lines = getMagicComments(bytes)
    regex = re.compile("coding[:=]\s*([-\w.]+)")
    for txt in lines:
        match = regex.search(txt)
        if match:
            log.debug("guessEncoding: Found encoding %s" % match.group(1))
            return match.group(1), None
    return None, None


def encodedBytesToUnicode(bytes, encoding, bom):
    if bom:
        start = len(bom)
    else:
        start = 0
    unicodestring = bytes[start:].decode(encoding)
    return unicodestring


def parseEmacs(header):
    """Determine if the header specifies a major mode.
    
    Parse a potential emacs major mode specifier line into the
    mode and the optional variables.  The mode may appears as any
    of::

      -*-C++-*-
      -*- mode: Python; -*-
      -*- mode: Ksh; var1:value1; var3:value9; -*-

    @param header: first x bytes of the file to be loaded
    @return: two-tuple of the mode and a dict of the name/value pairs.
    @rtype: tuple
    """
    import emacsutil

    lines = getMagicComments(header)
    for line in lines:
        mode, vars = emacsutil.parseModeline(line)
        if mode:
            return mode, vars
    return None, None


def guessBinary(text, percentage=5):
    """Guess if this is a text or binary file.
    
    Guess if the text in this file is binary or text by scanning
    through the first C{amount} characters in the file and
    checking if some C{percentage} is out of the printable ascii
    range.

    Obviously this is a poor check for unicode files, so this is
    just a bit of a hack.

    @param amount: number of characters to check at the beginning
    of the file

    @type amount: int

    @param percentage: percentage of characters that must be in
    the printable ASCII range

    @type percentage: number

    @rtype: boolean
    """
    encoding, bom = detectEncoding(text)
    if encoding:
        # The presence of an encoding by definition indicates a text file, so
        # therefore not binary!
        return False
    data = [ord(i) for i in text]
    binary=0
    for ch in data:
        if (ch<8) or (ch>13 and ch<32) or (ch>126):
            binary+=1
    log.debug("guessBinary: len=%d, num binary=%d" % (len(text), binary))
    if binary>(len(text)/percentage):
        return True
    return False


def guessSpacesPerIndent(text):
    """Guess the number of spaces per indent level
    
    Takes from the SciTE source file SciTEBase::DiscoverIndentSetting
    """
    tabsizes = [0]*9
    indent = 0 # current line indentation
    previndent = 0 # previous line indentation
    prevsize = -1 # previous line tab size
    newline = True
    for c in text:
        if c == '\n' or c == '\r':
            newline = True
            indent = 0
        elif newline:
            if c == ' ':
                indent += 1
            else:
                if indent:
                    if indent == previndent and prevsize >= 0:
                        tabsizes[prevsize] += 1
                    elif indent > previndent and previndent >= 0:
                        if indent - previndent <= 8:
                            prevsize = indent - previndent
                            tabsizes[prevsize] += 1
                        else:
                            prevsize = -1
                    previndent = indent
                elif c == '\t':
                    tabsizes[0] += 1
                newline = False

    # find maximum non-zero indent
    index = -1
    for i, size in enumerate(tabsizes):
        if size > 0 and (index == -1 or size > tabsizes[index]):
            index = i

    return index

# HTML converters from Chris Barker:
#
# This is an alternative to using <pre> -- I want it to wrap, but otherwise
# preserve newlines, etc.  Perhaps is would make sense to use some of
# the stuff ion webhelpers, instead of writting this all from scratch:
# http://webhelpers.groovie.org/


def wrapFont(text, font_size=None):
    if font_size is not None:
        text = "<font size=%d>%s</font>" % (font_size, text)
    return text


def text2HtmlPlain(text, font_size=None):
    """
    Takes raw text, and returns html with newlines converted, etc.
    
    Used the default font, with no extra whitespace games.
    """
    # double returns are a new paragraph
    text = text.split("\n\n")
    text = [p.strip().replace("\n","<br>") for p in text]
    body = wrapFont("<p>\n" + "\n</p>\n<p>\n".join(text) + "\n</p>\n", font_size)
    text = "<html>\n<body>\n" + body + "</body>\n</html>"
    return text


def text2HtmlFixed(text, font_size=None):
    """
    Takes raw text, and returns html with newlines converted, etc., using a fixed width font.
    It should also preserve text lined up with spaces.
    """
    text = text.replace("\n","<br>")
    # this preserves multiple spaces: it does mess up breaking on a two space gap
    text = text.replace("  ","&nbsp;&nbsp;")
    body = wrapFont("<p>\n" + text + "\n</p>\n", font_size)
    text = "<html>\n<body>\n<tt>\n" + body + "</tt></body>\n</html>"
    return text


def text2HtmlParagraph(text, font_size=None):
    """
    Takes raw text, and returns html with consecutive non-blank lines
    converted to paragraphs.
    
    Used the default font, with no extra whitespace games.
    """
    # double returns are a new paragraph
    text = text.split("\n\n")
    text = [p.strip().replace("\n","") for p in text]
    body = wrapFont("<p>\n" + "\n</p>\n<p>\n".join(text) + "\n</p>\n", font_size)
    text = "<html>\n<body>\n" + body + "</body>\n</html>"
    return text


def text_to_int(text, default_base="dec"):
    """ Convert text to int, raising exeception on invalid input
    """
    if text.startswith("0x"):
        value = int(text[2:], 16)
    elif text.startswith("$"):
        value = int(text[1:], 16)
    elif text.startswith("#"):
        value = int(text[1:], 10)
    elif text.startswith("%"):
        value = int(text[1:], 2)
    else:
        if default_base == "dec":
            value = int(text)
        else:
            value = int(text, 16)
    return value


def parse_int_label_dict(text, base=10, allow_equals=False):
    """Parse a multi-line text string into a dict keyed on integers with a
    keyword value

    For each line, it tries to find the first integer-looking thing and use
    that as the key, and the first label-looking thing (alphanumeric, starting
    with a letter or _, and stopping at the first non-alphanumeric or
    whitespace) and use that as the label.

    If ``allow_equal`` is True, it will check for the presense of an ``=``
    character and parse in the reverse order, where the label comes first and
    the address follows after the equals character.
    """
    regex = re.compile("(0x|\$)?([0-9a-fA-F]+)[^a-zA-Z]+([a-zA-Z_]\w*)")
    regex_eq = re.compile("([a-zA-Z_]\w*)[^a-zA-Z=]*=.*?(0x|\$)?([0-9a-fA-F]+)")
    d = {}
    for line in text.splitlines():
        match = None
        if allow_equals and "=" in line:
            match = regex_eq.search(line)
            key = match.group(3)
            value = match.group(1)
            b = match.group(2)

        if not match:
            match = regex.search(line)

            if match:
                key = match.group(2)
                value = match.group(3)
                b = match.group(1)
        if match:
            line_base = 16 if b == "$" or b == "0x" else 10 if b == "#" else base
            try:
                key = int(key, line_base)
                d[key] = value
            except ValueError:
                # just ignore bad lines
                pass
    return d


def slugify(s):
    """Simplifies ugly strings into something URL-friendly.

    >>> print slugify("[Some] _ Article's Title--")
    some-articles-title

    From https://gist.github.com/dolph/3622892#file-slugify-py
    """

    # "[Some] _ Article's Title--"
    # "[some] _ article's title--"
    s = s.lower()

    # "[some] _ article's_title--"
    # "[some]___article's_title__"
    for c in [' ', '-', '.', '/']:
        s = s.replace(c, '_')

    # "[some]___article's_title__"
    # "some___articles_title__"
    s = re.sub('\W', '', s)

    # "some___articles_title__"
    # "some   articles title  "
    s = s.replace('_', ' ')

    # "some   articles title  "
    # "some articles title "
    s = re.sub('\s+', ' ', s)

    # "some articles title "
    # "some articles title"
    s = s.strip()

    # "some articles title"
    # "some-articles-title"
    s = s.replace(' ', '-')

    return s


if __name__ == "__main__":
    import sys

    for file in sys.argv[1:]:
        fh = open(file)
        text = fh.read()
        log.debug("file=%s, tabsize=%d" % (file, guessSpacesPerIndent(text)))

        print parse_int_label_dict(text)
        print parse_int_label_dict(text, allow_equal=True)



#------------------------------------------------------------------------------
#
#  Copyright (c) 2009, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: Evan Patterson
#  Date:   06/18/2009
#
#------------------------------------------------------------------------------

""" Provides a lightweight framework that removes some of the drudge work
    involved with implementing user-friendly saving behavior in a Traits
    UI application.
"""

from __future__ import absolute_import

# ETS imports
from pyface.api import FileDialog, confirm, error, YES, CANCEL
from pyface.timer.api import Timer
from traits.api import HasTraits, Str, Bool, Any, Int, Instance, on_trait_change
from ..api import Handler


class CanSaveMixin(HasTraits):
    """ A mixin-class for objects that wish to support GUI saving via a
        SaveHandler. It is the responsiblity of the child class to manage
        its dirty flag, which describes whether its information has changed
        since its last save.
    """

    filepath = Str
    dirty = Bool(False)

    #-----------------------------------------------------------------
    #  object interface
    #-----------------------------------------------------------------

    def __getstate__(self):
        """ We don't want to pickle the filepath because this can change,
            obviously, if the user moves around the pickled file.
        """
        state = super(CanSaveMixin, self).__getstate__()
        del state['filepath']
        del state['dirty']
        return state

    #-----------------------------------------------------------------
    #  CanSaveMixin interface
    #-----------------------------------------------------------------

    def validate(self):
        """ Returns whether the information in the object is valid to be saved
            in tuple form. The first item is the validation state (boolean) and
            the second item is the message to display if the object did not
            validate.

            By default, an object always validates.
        """
        return (True, '')

    def save(self):
        """ Saves the object to the path specified by its 'filepath' trait. This
            method should also reset the dirty flag on this object.
        """
        raise NotImplementedError


class SaveHandler(Handler):
    """ A Handler that facilates adding saving to a Traits UI application.
    """

    # The object which is to be saved (subclass of CanSaveMixin). It is assigned
    # to info.object in the 'init' method, which in many cases is what you want.
    # If not, override that method to set it to something else.
    saveObject = Any

    # The type of files to show in the save dialogs
    wildcard = Str('All files (*.*)|*.*')

    # The option extension which should appear at the end of all filenames. If
    # the user does not explicitly specifiy it, it is appended to the filename.
    extension = Str

    # This message to display when the Handler requests a save
    savePromptMessage = Str('Would you like to save?')

    # Whether to prompt for a save on exit if the save object is dirty
    promptOnExit = Bool(True)

    # Whether to allow the user to override a validation failure through a
    # confirmation dialog. By default, validation errors cannot be overriden.
    allowValidationBypass = Bool(False)

    # Whether to automatically save after a certain amount of time has passed
    # since the last save
    autosave = Bool(False)

    # Number of seconds between each autosave. Default is 5 minutes.
    autosaveInterval = Int(300)

    # If it is possible to override validation failures, this specifies whether
    # autosave will do so. If False and a validation errors occurs, no save
    # will occur.
    autosaveValidationBypass = Bool(True)

    # Protected traits
    _timer = Instance(Timer)

    #-----------------------------------------------------------------
    #  Handler interface
    #-----------------------------------------------------------------

    def init(self, info):
        """ Set the default save object (the object being handled). Also,
            perform a questionable hack by which we remove the handled
            object from the keybinding's controllers. This means that a
            keybinding to 'save' only calls this object, not the object
            being edited as well. (For reasons unclear, the KeyBinding handler
            API is radically different from the Action API, which is the reason
            that this problem exists. Keybindings are a UI concept--they should
            *not* call the model by default.)
        """
        keybindings = info.ui.key_bindings
        if keybindings is not None:
            keybindings.controllers.remove(info.object)

        self.saveObject = info.object
        return True

    def close(self, info, is_ok):
        """ Called when the user requests to close the interface. Returns a
            boolean indicating whether the window should be allowed to close.
        """
        if self.promptOnExit:
            return self.promptForSave(info)
        else:
            return True

    def closed(self, info, is_ok):
        """ Called after the window is destroyed. Makes sure that the autosave
            timer is stopped.
        """
        if self._timer:
            self._timer.Stop()

    #-----------------------------------------------------------------
    #  SaveHandler interface
    #-----------------------------------------------------------------

    def exit(self, info):
        """ Closes the UI unless a save prompt is cancelled. Provided for
            convenience to be used with a Menu action.
        """
        if self.close(info, True):
            info.ui.dispose()

    def save(self, info):
        """ Saves the object to its current filepath. If this is not specified,
            opens a dialog to select this path. Returns whether the save
            actually occurred.
        """
        if self.saveObject.filepath == '':
            return self.saveAs(info)
        else:
            return self._validateAndSave()

    def saveAs(self, info):
        """ Saves the object to a new path, and sets this as the 'filepath' on
            the object. Returns whether the save actually occurred.
        """
        fileDialog = FileDialog(action='save as', title='Save As',
                                wildcard=self.wildcard,
                                parent=info.ui.control)
        fileDialog.open()
        if fileDialog.path == '' or fileDialog.return_code == CANCEL:
            return False
        else:
            extLen = len(self.extension)
            if extLen and fileDialog.path[-extLen -
                                          1:] != '.' + self.extension:
                fileDialog.path += '.' + self.extension
            self.saveObject.filepath = fileDialog.path
            return self._validateAndSave()

    def promptForSave(self, info, cancel=True):
        """ Prompts the user to save the object, if appropriate. Returns whether
            the user canceled the action that invoked this prompt.
        """
        if self.saveObject.dirty:
            code = confirm(info.ui.control, self.savePromptMessage,
                           title="Save now?", cancel=cancel)
            if code == CANCEL:
                return False
            elif code == YES:
                if not self.save(info):
                    return self.promptForSave(info, cancel)
        return True

    def _autosave(self):
        """ Called by the timer when an autosave should take place.
        """
        if self.saveObject.dirty and self.saveObject.filepath != '':
            success, message = self.saveObject.validate()
            if success or (self.allowValidationBypass and
                           self.autosaveValidationBypass):
                self.saveObject.save()

    @on_trait_change('autosave, autosaveInterval, saveObject')
    def _configure_timer(self):
        """ Creates, replaces, or destroys the autosave timer.
        """
        if self._timer:
            self._timer.Stop()
        if self.autosave and self.saveObject:
            self._timer = Timer(self.autosaveInterval * 1000, self._autosave)
        else:
            self._timer = None

    def _validateAndSave(self):
        """ Try to save to the current filepath. Returns whether whether the
            validation was successful/overridden (and the object saved).
        """
        success, message = self.saveObject.validate()
        if success:
            self.saveObject.save()
        else:
            title = "Validation error"
            if (self.allowValidationBypass and
                    confirm(None, message, title=title) == YES):
                self.saveObject.save()
                success = True
            else:
                error(None, message, title=title)
        return success



#------------------------------------------------------------------------------
# Copyright (c) 2008, Riverbank Computing Limited
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Riverbank Computing Limited
# Description: <Enthought permissions package component>
#------------------------------------------------------------------------------


# Enthought library imports.
from traits.api import Bool, Instance, Interface

# Local imports.
from i_user_storage import IUserStorage


class IUserDatabase(Interface):
    """The interface to be implemented by a user database for the default user
    manager."""

    # Set if the implementation supports changing a user's password.
    can_change_password = Bool

    # Set if the implementation supports adding users.
    can_add_user = Bool

    # Set if the implementation supports modifying users.
    can_modify_user = Bool

    # Set if the implementation supports deleting users.
    can_delete_user = Bool

    # The user data storage.
    user_storage = Instance(IUserStorage)

    def bootstrapping(self):
        """Return True if the user database is bootstrapping.  Typically this
        is when no users have been defined."""

    def authenticate_user(self, user):
        """Authenticate the given user and return True if successful.  user
        implements IUser."""

    def unauthenticate_user(self, user):
        """Unauthenticate the given user and return True if successful.  user
        implements IUser."""

    def change_password(self, user):
        """Change a user's password in the database.  This only needs to be
        reimplemented if 'can_change_password' is True."""

    def add_user(self):
        """Add a user account to the database.  This only needs to be
        reimplemented if 'can_add_user' is True."""

    def modify_user(self):
        """Modify a user account in the database.  This only needs to be
        reimplemented if 'can_modify_user' is True."""

    def delete_user(self):
        """Delete a user account from the database.  This only needs to be
        reimplemented if 'can_delete_user' is True."""

    def matching_user(self, name):
        """Return an object that implements IUser for the user selected based
        on the given name.  If there was no user selected then return None.
        How the name is interpreted (eg. as a regular expression) is determined
        by the user database.  Note that the blob attribute of the user will
        not be set."""

    def user_factory(self):
        """Return a new object that implements the IUser interface."""



#-------------------------------------------------------------------------
#
#  Copyright (c) 2007, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   07/03/2007
#
#-------------------------------------------------------------------------

""" Defines 'ThemedControl, a themed control based class. A 'themed' control is
    a control (optionally) supporting a stretchable background image.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

import wx

from traits.api \
    import HasTraits, Str, Int, Enum, Bool, Property, Event, Tuple, Instance, \
    cached_property, on_trait_change

from traitsui.api \
    import default_theme

from traitsui.ui_traits \
    import Image, Position, Alignment, Spacing

from image_slice import ImageSlice

from themed_window \
    import ThemedWindow

#-------------------------------------------------------------------------
#  Constants:
#-------------------------------------------------------------------------

# The size of an empty text string:
ZeroTextSize = (0, 0, 0, 0)

# An empty position and size bounds:
EmptyBounds = (0, 0, 0, 0)

# Targets for '_get_bounds_for' method:
TheText = 0
TheBitmap = 1
TheControl = 2

#-------------------------------------------------------------------------
#  'ThemedControl' class:
#-------------------------------------------------------------------------


class ThemedControl(ThemedWindow):

    #-- Public Traits --------------------------------------------------------

    # An (optional) image to be drawn inside the control:
    image = Image(event='updated')

    # The (optional) text to be displayed inside the control:
    text = Str(event='updated')

    # The position of the image relative to the text:
    position = Position(event='updated')

    # The amount of spacing between the image and the text:
    spacing = Spacing(event='updated')

    # Is the text value private (like a password):
    password = Bool(False, event='updated')

    # An additional, optional offset to apply to the text/image position:
    offset = Tuple(Int, Int)

    # Minimum default size for the control:
    min_size = Tuple(Int, Int)

    # Is the control enabled:
    enabled = Bool(True)

    #-- Private Traits -------------------------------------------------------

    # An event fired when any display related value changes:
    updated = Event

    # The underlying wx.Window control:
    control = Instance(wx.Window)

    # The current text value to display:
    current_text = Property(depends_on='text, password')

    # The best size for the control:
    best_size = Property

    # The size of the current text:
    text_size = Property(depends_on='current_text, control')

    # The position and size of the current text within the control:
    text_bounds = Property(depends_on='updated')

    # The position and size of the current bitmap within the control:
    bitmap_bounds = Property(depends_on='updated')

    #-- Public Methods -------------------------------------------------------

    def create_control(self, parent):
        """ Creates the underlying wx.Window object.
        """
        self.control = control = wx.Window(parent, -1,
                                           size=wx.Size(70, 20),
                                           style=wx.FULL_REPAINT_ON_RESIZE | wx.WANTS_CHARS)

        # Initialize the control (set-up event handlers, ...):
        self.init_control()

        # Make sure all internal state gets initialized:
        self._image_changed(self.image)

        # Make sure the control is sized correctly:
        size = self.best_size
        control.SetMinSize(size)
        control.SetSize(size)

        return control

    #-- Property Implementations ---------------------------------------------

    @cached_property
    def _get_current_text(self):
        """ Returns the current text to display.
        """
        if self.password:
            return '*' * len(self.text)

        return self.text

    def _get_best_size(self):
        """ Returns the 'best' size for the control.
        """
        cx, cy, cdx, cdy = self._get_bounds_for(TheControl)
        mdx, mdy = self.min_size
        return wx.Size(max(cdx, mdx), max(cdy, mdy))

    @cached_property
    def _get_text_size(self):
        """ Returns the text size information for the window.
        """
        text = self.current_text
        if (text == '') or (self.control is None):
            return ZeroTextSize

        return self.control.GetFullTextExtent(text)

    @cached_property
    def _get_text_bounds(self):
        """ Returns the position and size of the text within the window.
        """
        return self._get_bounds_for(TheText)

    @cached_property
    def _get_bitmap_bounds(self):
        """ Returns the size and position of the bitmap within the window.
        """
        return self._get_bounds_for(TheBitmap)

    #-- Event Handlers -------------------------------------------------------

    @on_trait_change('theme.+, image')
    def _updated_changed(self):
        """ Handles any update related trait being changed.
        """
        if self.control is not None:
            self.control.Refresh()

    def _image_changed(self, image):
        """ Handles the image being changed by updating the corresponding
            bitmap.
        """
        if image is None:
            self._bitmap = None
        else:
            self._bitmap = image.create_image().ConvertToBitmap()

    #-- wxPython Event Handlers ----------------------------------------------

    def _paint_fg(self, dc):
        """ Paints the foreground into the specified device context.
        """
        self.enabled = self.control.IsEnabled()

        # Get the text and image offset to use:
        theme = self.theme or default_theme
        label = theme.label
        ox, oy = label.left, label.top
        ox2, oy2 = self.offset
        ox += ox2
        oy += oy2

        # Draw the bitmap (if any):
        ix, iy, idx, idy = self.bitmap_bounds
        if idx != 0:
            dc.DrawBitmap(self._bitmap, ix + ox, iy + oy, True)

        # Draw the text (if any):
        tx, ty, tdx, tdy = self.text_bounds
        if tdx != 0:
            dc.SetBackgroundMode(wx.TRANSPARENT)
            dc.SetTextForeground(theme.content_color)
            dc.SetFont(self.control.GetFont())
            dc.DrawText(self.current_text, tx + ox, ty + oy)

    def _size(self, event):
        """ Handles the control being resized.
        """
        super(ThemedControl, self)._size(event)
        self.updated = True

    #-- Private Methods ------------------------------------------------------

    def _get_bounds_for(self, item):
        """ Returns all text and image related position and size information.
        """
        control = self.control
        if control is None:
            return EmptyBounds

        tdx, tdy, descent, leading = self.text_size
        bitmap = self._bitmap
        if bitmap is None:
            bdx, bdy = 0, 0
        else:
            bdx, bdy = bitmap.GetWidth(), bitmap.GetHeight()
        if (tdx + bdx) == 0:
            return EmptyBounds

        wdx, wdy = control.GetClientSizeTuple()
        spacing = (tdx != 0) * (bdx != 0) * self.spacing
        theme = self.theme or default_theme
        slice = theme.image_slice or ImageSlice()
        content = theme.content

        position = self.position
        if position in ('above', 'below'):
            cdx = max(tdx, bdx)
            cdy = tdy + spacing + bdy
        else:
            cdx = tdx + spacing + bdx
            cdy = max(tdy, bdy)

        if item == TheControl:
            cdx += content.left + content.right
            cdy += content.top + content.bottom

            return (0, 0, max(slice.left + slice.right,
                              slice.xleft + slice.xright + cdx),
                    max(slice.top + slice.bottom,
                        slice.xtop + slice.xbottom + cdy))

        alignment = theme.alignment
        if alignment == 'default':
            alignment = self.default_alignment
        if alignment == 'left':
            x = slice.xleft + content.left
        elif alignment == 'center':
            x = slice.xleft + content.left + ((wdx - slice.xleft - slice.xright
                                               - content.left - content.right - cdx) / 2)
        else:
            x = wdx - slice.xright - content.right - cdx

        if position == 'left':
            bx = x
            tx = bx + bdx + spacing
        elif position == 'right':
            tx = x
            bx = tx + tdx + spacing
        else:
            x += (max(tdx, bdx) / 2)
            tx = x - (tdx / 2)
            bx = x - (bdx / 2)

        y = slice.xtop + content.top + ((wdy - slice.xtop - slice.xbottom -
                                         content.top - content.bottom - cdy) / 2)
        if position == 'above':
            by = y
            ty = by + bdy + spacing
        elif position == 'below':
            ty = y
            by = ty + tdy + spacing
        else:
            y += (max(tdy, bdy) / 2)
            ty = y - ((tdy + 1) / 2)
            by = y - (bdy / 2)

        if item == TheText:
            return (tx, ty, tdx, tdy)

        return (bx, by, bdx, bdy)



#------------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: Enthought, Inc.
#
#------------------------------------------------------------------------------

""" A status bar manager realizes itself in a status bar control.
"""

# Major package imports.
import wx

# Enthought library imports.
from traits.api import Any, HasTraits, List, Property, Str, Unicode


class StatusBarManager(HasTraits):
    """ A status bar manager realizes itself in a status bar control. """

    # The message displayed in the first field of the status bar.
    message = Property

    # The messages to be displayed in the status bar fields.
    messages = List(Unicode)

    # The toolkit-specific control that represents the status bar.
    status_bar = Any

    ###########################################################################
    # 'StatusBarManager' interface.
    ###########################################################################

    def create_status_bar(self, parent):
        """ Creates a status bar. """

        if self.status_bar is None:
            self.status_bar = wx.StatusBar(parent)
            self.status_bar._pyface_control = self
            if len(self.messages) > 1:
                self.status_bar.SetFieldsCount(len(self.messages))
                for i in range(len(self.messages)):
                    self.status_bar.SetStatusText(self.messages[i], i)
            else:
                self.status_bar.SetStatusText(self.message)

        return self.status_bar

    def remove_status_bar(self, parent):
        """ Removes a status bar. """

        if self.status_bar is not None:
            self.status_bar.Destroy()
            self.status_bar._pyface_control = None
            self.status_bar = None

    ###########################################################################
    # Property handlers.
    ###########################################################################

    def _get_message(self):
        """ Property getter. """

        if len(self.messages) > 0:
            message = self.messages[0]

        else:
            message = ''

        return message

    def _set_message(self, value):
        """ Property setter. """

        if len(self.messages) > 0:
            old = self.messages[0]
            self.messages[0] = value

        else:
            old = ''
            self.messages.append(value)

        self.trait_property_changed('message', old, value)

        return

    ###########################################################################
    # Trait event handlers.
    ###########################################################################

    def _messages_changed(self):
        """ Sets the text displayed on the status bar. """

        if self.status_bar is not None:
            for i in range(len(self.messages)):
                self.status_bar.SetStatusText(self.messages[i], i)

        return

    def _messages_items_changed(self):
        """ Sets the text displayed on the status bar. """

        if self.status_bar is not None:
            for i in range(len(self.messages)):
                self.status_bar.SetStatusText(self.messages[i], i)

        return

#### EOF ######################################################################



#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought logger package component>
#------------------------------------------------------------------------------

# Standard library imports.
import logging
import os

# Enthought library imports.
from traits.util.home_directory import get_home_directory


# Setup a logger for this module.
logger = logging.getLogger(__name__)


def create_email_message(fromaddr, toaddrs, ccaddrs, subject, priority,
                         include_project=False, stack_trace="", comments=""):
    # format a message suitable to be sent to the Roundup bug tracker

    from email.MIMEMultipart import MIMEMultipart
    from email.MIMEText import MIMEText
    from email.MIMEBase import MIMEBase

    message = MIMEMultipart()
    message['Subject'] = "%s [priority=%s]" % (subject, priority)
    message['To'] = ', '.join(toaddrs)
    message['Cc'] = ', '.join(ccaddrs)
    message['From'] = fromaddr
    message.preamble = 'You will not see this in a MIME-aware mail reader.\n'
    message.epilogue = ' ' # To guarantee the message ends with a newline

    # First section is simple ASCII data ...
    m = []
    m.append("Bug Report")
    m.append("==============================")
    m.append("")

    if len(comments) > 0:
        m.append("Comments:")
        m.append("========")
        m.append(comments)
        m.append("")

    if len(stack_trace) > 0:
        m.append("Stack Trace:")
        m.append("===========")
        m.append(stack_trace)
        m.append("")

    msg = MIMEText('\n'.join(m))
    message.attach(msg)

    # Include the log file ...
    if True:
        try:
            log = os.path.join(get_home_directory(), 'envisage.log')
            f = open(log, 'r')
            entries = f.readlines()
            f.close()

            ctype = 'application/octet-stream'
            maintype, subtype = ctype.split('/', 1)
            msg = MIMEBase(maintype, subtype)

            msg = MIMEText(''.join(entries))
            msg.add_header('Content-Disposition', 'attachment', filename='logfile.txt')
            message.attach(msg)
        except:
            logger.exception('Failed to include log file with message')

    # Include the environment variables ...
    if True:
        """
        Transmit the user's environment settings as well.  Main purpose is to
        work out the user name to help with following up on bug reports and
        in future we should probably send less data.
        """
        try:
            entries = []
            for key, value in os.environ.iteritems():
                entries.append('%30s : %s\n' % (key, value))

            ctype = 'application/octet-stream'
            maintype, subtype = ctype.split('/', 1)
            msg = MIMEBase(maintype, subtype)

            msg = MIMEText(''.join(entries))
            msg.add_header('Content-Disposition', 'attachment', filename='environment.txt')
            message.attach(msg)

        except:
            logger.exception('Failed to include environment variables with message')


# FIXME: no project plugins exist for Envisage 3, yet, and this isn't the right
# way to do it, either. See the docstring of attachments.py.
#    # Attach the project if requested ...
#    if include_project:
#        from attachments import Attachments
#        try:
#            attachments = Attachments(message)
#            attachments.package_any_relevant_files()
#        except:
#            logger.exception('Failed to include workspace files with message')

    return message


